//===========================================================
// WINCMD - A command language interpeter for Windows
// Copyright (C) 1993 Douglas Boling
//
// Revision History:
//
// 1.0   Inital Release
//
// 1.1   Fixed AppendString bug
//       Fixed Lib load directory find
//       Fixed Cmp Op assignment bug
//       Fixed Launch bug
//       Removed trailing space from ResolveLine code
//       Allowed \ and : in variable names
//       Added WINCMDVER predefined var
//       Added Library About menus
//       Added Help screen
//
// 1.2   Fixed end of file problem causing GPFs
//       Added hooks to allow statements in libraries
//       Added hooks for library provided error messages
//       Added hooks for ON type statements 
//
// 1.3   Fixed SendKey F10 problem. 
//       Fixed problem with signed compare functions
//       Cleaned up termination code 
//       Fixed library statement hooks
//       Fixed Stop statement asych problem
//       Fixed incomplete reset problem
//       Added LaunchSpecial statement
//       Added GetDropCnt and GetDropFile functions
// Else not on next line problem?
//
// 1.4   Fixed stupid yield problem
//
//===========================================================
// Returns no. of elements
#define dim(x) (sizeof(x) / sizeof(x[0]))   

#define	VARBUFFSIZE              42000
#define	TOKBUFFSIZE              65500

#define	MAXLINELEN               256
#define	MAXSTRLEN                252
#define	MAXOUTLINES              100

#define	MAXLIBS                  32
#define	MAXLIBERRS               512 
//-----------------------------------------------------------
// Include flags for WINDOWS.H
//-----------------------------------------------------------
#define NOCOMM             1    Comm driver APIs and definitions
#define NOMINMAX           1    min() and max() macros
#define NOLOGERROR         1    LogError() and related definitions
#define NOPROFILER         1    Profiler APIs
#define NORESOURCE         1    Resource management
#define NOATOM             1    Atom management
#define NOLANGUAGE         1    Character test routines
#define NODBCS             1    Double-byte character set routines
#define NOGDICAPMASKS      1    GDI device capability constants
//#define NODRAWTEXT         1    DrawText() and related definitions
#define NOTEXTMETRIC       1    TEXTMETRIC and related APIs
#define NOSCALABLEFONT     1    Truetype scalable font support
#define NOBITMAP           1    Bitmap support
#define NORASTEROPS        1    GDI Raster operation definitions
#define NOMETAFILE         1    Metafile support
#define NOSYSTEMPARAMSINFO 1    SystemParametersInfo() and SPI_#define definitions
#define NODEFERWINDOWPOS   1    DeferWindowPos and related definitions
#define NOKEYSTATES        1    MK_* message key state flags
#define NOWH               1    SetWindowsHook and related WH_* definitions
#define NOSCROLL           1    Scrolling APIs and scroll bar control
//#define NOVIRTUALKEYCODES  1    VK_* virtual key codes
//#define NOMENUS            1    Menu APIs
//#define NOKEYBOARDINFO     1    Keyboard driver routines
//#define NOCOLOR            1    COLOR_#define color values
//#define NOGDIOBJ           1    GDI pens, brushes, fonts
//#define NOMSG              1    APIs and definitions that use MSG structure
//#define NOWINSTYLES        1    Window style definitions

//-----------------------------------------------------------
// Include files
//-----------------------------------------------------------
#include "windows.h"
#include "commdlg.h"
#include "toolhelp.h"
#include "shellapi.h"

#include "stdlib.h"
#include "direct.h"
#include "dos.h"
#include "string.h"
#include "time.h"
#include "WinCmd.h"

// Message handler functions
LONG DoCreateMain (HWND, UINT, UINT, LONG);
LONG DoSizeMain (HWND, UINT, UINT, LONG);
LONG DoTimerMain (HWND, UINT, UINT, LONG);
LONG DoInitMenuMain (HWND, UINT, UINT, LONG);
LONG DoDropFilesMain (HWND, UINT, UINT, LONG);
LONG DoDestroyMain (HWND, UINT, UINT, LONG);
LONG DoCommandMain (HWND, UINT, UINT, LONG);
LONG DoRefreshMain (HWND, UINT, UINT, LONG);
LONG DoRestartMain (HWND, UINT, UINT, LONG);

// Menu Function Prototypes
LONG DoMainMenuFileOpen (HWND, UINT, HWND, UINT);
LONG DoMainMenuFileExit (HWND, UINT, HWND, UINT);
LONG DoMainMenuShowList (HWND, UINT, HWND, UINT);
LONG DoMainMenuShowOut (HWND, UINT, HWND, UINT);
LONG DoMainMenuClearOut (HWND, UINT, HWND, UINT);
LONG DoMainMenuMinOnRun (HWND, UINT, HWND, UINT);
LONG DoMainMenuHelp (HWND, UINT, HWND, UINT);
LONG DoMainMenuAbout (HWND, UINT, HWND, UINT);

// Control function Prototypes
LONG DoMainCtlRun (HWND, UINT, HWND, UINT);
LONG DoMainCtlStep (HWND, UINT, HWND, UINT);
LONG DoMainCtlReset (HWND, UINT, HWND, UINT);
// Utility function prototpes
BOOL MyYield (void);
BOOL MyGetFilename (HWND, char *, INT, char *, INT);
typedef struct tm TM;

typedef BOOL(WINAPI *THUREGPROC)(HANDLE);
typedef BOOL(WINAPI *THREGPROC)(HANDLE, FARPROC, WORD);
typedef BOOL(WINAPI *THTFH)(TASKENTRY far *, HTASK);
//------------------------------------------------------------
// Message dispatch table for MainWindowProc
//------------------------------------------------------------
struct decodeUINT MainMessages[] = {
	WM_COMMAND, DoCommandMain,
	WM_CREATE, DoCreateMain,
	WM_SIZE, DoSizeMain,
	WM_TIMER, DoTimerMain,
	WM_DROPFILES, DoDropFilesMain,
	WM_INITMENU, DoInitMenuMain,
	MYMSG_REFRESH, DoRefreshMain,
	MYMSG_RESTART, DoRestartMain,
	WM_DESTROY, DoDestroyMain
};

//------------------------------------------------------------
// Command Message dispatch for MainWindowProc
//------------------------------------------------------------
struct decodeCMD MainMenuItems[] = {
	IDD_RUN, DoMainCtlRun,
	IDD_STEP, DoMainCtlStep,
	IDD_RESET, DoMainCtlReset,
	IDM_OPEN, DoMainMenuFileOpen,
	IDM_EXIT, DoMainMenuFileExit,
	IDM_SHOWLIST, DoMainMenuShowList,
	IDM_SHOWOUT, DoMainMenuShowOut,		
	IDM_CLEAROUT, DoMainMenuClearOut,
	IDM_MINONRUN, DoMainMenuMinOnRun,
	IDM_HELP, DoMainMenuHelp,	
	IDM_ABOUT, DoMainMenuAbout
};

//-----------------------------------------------------------
// Application specific function prototypes
//-----------------------------------------------------------
INT InitInterpeter (HWND);
INT ResetInterpeter (HWND);
INT TermInterpeter (void);
INT ParseCmdLine (LPSTR);
void WriteOutput (HWND, LPSTR);
BOOL GetErrorString (INT, INT, char *, INT);
void PrintError (HWND, INT, INT, BOOL);
INT LoadProgram (HWND, LPLINETOKEN, UINT);
INT SetRunning (BOOL);
INT TokenizeLine (LPSTR *, LPLINETOKEN *, UINT);
INT AddTokenInit (char *, BYTE, UINT, DWORD);
PTOKEN AddToken (char *, BYTE, UINT, DWORD, INT *);
PTOKEN AddTokenL (LPSTR, BYTE, UINT, DWORD, LPINT);
INT DeleteToken (PTOKEN);
INT DeleteString (char *, WORD);
INT SetToken (PTOKEN, BYTE, UINT, DWORD);
PTOKEN FindToken (LPSTR, INT);
DWORD CALLBACK TokFuncCallback (INT, LPTOKEN, LPSTR, BYTE, UINT, DWORD);
LONG ComputeNum (LPSTR *, INT *);
INT GetOp (LPLINETOKEN *, PTOKEN, UINT *, INT);
INT Eval (LPLINETOKEN *, PTOKEN, INT);
LONG Eval2Num (LPLINETOKEN *, INT, INT *);
INT ExecuteLine (LPLINETOKEN *, INT);
INT Execute (LPLINETOKEN *, INT);
LONG GetTokenVal (PTOKEN);
char *GetTokenString (PTOKEN);
INT ResolveLine (LPLINETOKEN *, LPSTR, INT);
LONG LaunchString (LPSTR, UINT);
INT DropFiles (HWND, HANDLE);

INT ReadFile (char *, HANDLE);
INT StuffKeyString (char **);

INT LocalOpCPren (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpAdd (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpSub (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpMul (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpDiv (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpMod (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpCmp (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpNOT (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpNOTBits (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpAND (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpANDBits (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpOR (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpORBits (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpXOR (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpXORBits (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpLThan (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpLEThan (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpGThan (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpGEThan (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalOpNotEqual (LPLINETOKEN *, PTOKEN, PTOKEN);
INT LocalComma (LPLINETOKEN *, PTOKEN, PTOKEN);

INT LocalOpOPren (LPLINETOKEN *, PTOKEN);
INT LocalLENGTH (LPLINETOKEN *, PTOKEN);
INT LocalUPPER (LPLINETOKEN *, PTOKEN);
INT LocalSUBSTR (LPLINETOKEN *, PTOKEN);
INT LocalARG (LPLINETOKEN *, PTOKEN);
INT LocalWAIT (LPLINETOKEN *, PTOKEN);
INT LocalFindLib (LPLINETOKEN *, PTOKEN);
INT WinActivate (LPLINETOKEN *, PTOKEN);
INT WinSendKeys (LPLINETOKEN *, PTOKEN);
INT WinGetActWin (LPLINETOKEN *, PTOKEN);
INT WinGetNextWin (LPLINETOKEN *, PTOKEN);
INT WinGetWinEXE (LPLINETOKEN *, PTOKEN);
INT WinMsgBox (LPLINETOKEN *, PTOKEN);
INT WinAskBox (LPLINETOKEN *, PTOKEN);
INT WinTicks (LPLINETOKEN *, PTOKEN);
INT WinGetEnvVar (LPLINETOKEN *, PTOKEN);
INT WinGetDropFile (LPLINETOKEN *, PTOKEN);
INT WinGetDropCount (LPLINETOKEN *, PTOKEN);

INT LocalIF (LPLINETOKEN *, INT);
INT LocalThenElse (LPLINETOKEN *, INT);
INT LocalDO (LPLINETOKEN *, INT);
INT LocalEND (LPLINETOKEN *, INT);
INT LocalWHILE (LPLINETOKEN *, INT);
INT LocalSAY (LPLINETOKEN *, INT);
INT LocalLaunch (LPLINETOKEN *, INT);
INT LocalLaunchandWait (LPLINETOKEN *, INT);
INT LocalAssign (LPLINETOKEN *, INT);
INT LocalEXIT (LPLINETOKEN *, INT);
INT LocalSTOP (LPLINETOKEN *, INT);
INT LocalRETURN (LPLINETOKEN *, INT);
INT LocalLEAVE (LPLINETOKEN *, INT);
INT LocalChDir (LPLINETOKEN *, INT);
INT LocalCMT2EOL (LPLINETOKEN *, INT);
INT LocalOnDrop (LPLINETOKEN *, INT);
INT LocalCLEAROUTPUT (LPLINETOKEN *, INT);

//-----------------------------------------------------------
// Global data
//-----------------------------------------------------------
HANDLE	hInst;
HWND		hMain, hwndLList, hwndOList;
HWND		hwndStepBtn, hwndRunBtn, hwndResetBtn;
BOOL		bMinOnRun = FALSE;
BOOL		bShowOut = FALSE;
BOOL		bShowList = TRUE;
char		*pszAskAns;

char szAppName[] = "WinCmd";            // Application name
char szTitleText[] = "WinCmd Window";   // Window text
char szMenuName[] = "WinCmdMenu";       // Menu name
char szIconName[] = "WinCmdIcon";       // Icon name
char szProfileName[] = "WinCmd.ini";    // INI file name

LOCALOP OpList[] = {
//  Name     Type       Level  Function
	{")",   TTYPE_CPAREN,  1, LocalOpCPren},
	{"NOT", TTYPE_LUOP,   15, LocalOpNOT},
	{"NOTB",TTYPE_LUOP,   15, LocalOpNOTBits},
	{"*",   TTYPE_LOP,    14, LocalOpMul},
	{"/",   TTYPE_LOP,    14, LocalOpDiv},
	{"MOD", TTYPE_LOP,    14, LocalOpMod},
	{"+",   TTYPE_LOP,    13, LocalOpAdd},
	{"-",   TTYPE_LUOP,   13, LocalOpSub},
	{"<",   TTYPE_LOP,    11, LocalOpLThan},
	{"<=",  TTYPE_LOP,    11, LocalOpLEThan},
	{">",   TTYPE_LOP,    11, LocalOpGThan},
	{">=",  TTYPE_LOP,    11, LocalOpGEThan},
	{"=",   TTYPE_ASSIGN_CMP, 10, LocalOpCmp},
	{"==",  TTYPE_LOP,    10, LocalOpCmp},
	{"<>",  TTYPE_LOP,    10,  LocalOpNotEqual},
	{"ANDB",TTYPE_LOP,     9,  LocalOpANDBits},
	{"XORB",TTYPE_LOP,     8,  LocalOpXORBits},
	{"ORB", TTYPE_LOP,     7,  LocalOpORBits},
	{"AND", TTYPE_LOP,     6,  LocalOpAND},
	{"OR",  TTYPE_LOP,     5,  LocalOpOR},
	{"XOR", TTYPE_LOP,     5,  LocalOpXOR},
	{",",   TTYPE_BREAK,   1,  LocalComma},
};

LOCALFUNC FuncList[] = {
//   Name         Type             Function
	{"(",         TTYPE_OPAREN,    LocalOpOPren},
	{"LENGTH",    TTYPE_LOCALFUNC, LocalLENGTH},
	{"UCASE",     TTYPE_LOCALFUNC, LocalUPPER},
	{"SUBSTR",    TTYPE_LOCALFUNC, LocalSUBSTR},
	{"ARG",       TTYPE_LOCALFUNC, LocalARG},
	{"DELAY",     TTYPE_LOCALFUNC, LocalWAIT},
	{"LIBVER",    TTYPE_LOCALFUNC, LocalFindLib},
	{"APPACTIVATE", TTYPE_LOCALFUNC, WinActivate},
	{"GETAPPACTIVE", TTYPE_LOCALFUNC, WinGetActWin},
	{"GETAPPEXE", TTYPE_LOCALFUNC, WinGetWinEXE},
	{"GETNEXTAPP", TTYPE_LOCALFUNC, WinGetNextWin},
	{"SENDKEYS",  TTYPE_LOCALFUNC, WinSendKeys},
	{"MSGBOX",    TTYPE_LOCALFUNC, WinMsgBox},
	{"ASKBOX",    TTYPE_LOCALFUNC, WinAskBox},
	{"TICKS",     TTYPE_LOCALFUNC, WinTicks},
	{"GETENVVAR", TTYPE_LOCALFUNC, WinGetEnvVar},
	{"GETDROPFILE", TTYPE_LOCALFUNC, WinGetDropFile},
	{"GETDROPCOUNT", TTYPE_LOCALFUNC, WinGetDropCount},
};

LOCALSTATEMENT StatementList[] = {
//   Name           Type          Function
	{"ONDROP",  TTYPE_LSTATEMENT, LocalOnDrop},
	{"LAUNCHSPECIAL", TTYPE_LSTATEMENT, LocalLaunchandWait},
	{"CLEAROUTPUT", TTYPE_LSTATEMENT, LocalCLEAROUTPUT},
	{"WHILE",   TTYPE_LSTATEMENT, LocalWHILE},
	{"IF",      TTYPE_LSTATEMENT, LocalIF},
	{"THEN",    TTYPE_THENELSE,   LocalThenElse},
	{"ELSE",    TTYPE_THENELSE,   LocalThenElse},
	{"DO",      TTYPE_LSTATEMENT, LocalDO},
	{"END",     TTYPE_LSTATEMENT, LocalEND},
	{"EXIT",    TTYPE_LSTATEMENT, LocalEXIT},
	{"STOP",    TTYPE_LSTATEMENT, LocalSTOP},
	{"RETURN",  TTYPE_LSTATEMENT, LocalRETURN},
	{"LEAVE",   TTYPE_LSTATEMENT, LocalLEAVE},
	{"PRINT",   TTYPE_LSTATEMENT, LocalSAY},
	{"SAY",     TTYPE_LSTATEMENT, LocalSAY},
	{"CD",      TTYPE_LSTATEMENT, LocalChDir},
	{"CHDIR",   TTYPE_LSTATEMENT, LocalChDir},
	{"REM",     TTYPE_CMT2EOL,    LocalCMT2EOL},
	{"//",      TTYPE_CMT2EOL,    LocalCMT2EOL},
	{"'",       TTYPE_CMT2EOL,    LocalCMT2EOL},
};
// Contstant tokens
struct {
	char *szName;
	LONG lData;
} ConstList[] = {{"HINST", 0}, {"HMAIN", 0}, {"WINCMDVER", 0},
                 {"MB_OK", MB_OK}, {"MB_OKCANCEL", MB_OKCANCEL}, 
                 {"MB_YESNO", MB_YESNO}, {"MB_YESNOCANCEL", MB_YESNOCANCEL},
					  {"MB_ICONQUESTION", MB_ICONQUESTION}, 
					  {"MB_ICONSTOP", MB_ICONSTOP}, {"ANS_CANCEL", IDCANCEL},
					  {"ANS_YES", IDYES}, {"ANS_NO", IDNO}, {"ANS_OK", IDOK},
					  {"RETURNCODE", 0x10000}, {"_DROPPTR", 0}, 
					  {"LC_HIDE", LC_HIDE}, {"LC_NORMAL", LC_SHOW}, {"LC_WAIT", LC_WAIT}, 
					  {"LC_MINIMIZE", LC_ICON}, {"LC_BACKGROUND", LC_BACKGROUND},
					  {"LC_MAXIMIZE", LC_MAXIMIZE},
                };
//
// Table for DOS Internal Commands
//
struct {
   char	*szName;
} CmdList[] = {             //DOS internal commands
	{"MD"}, {"RD"}, {"CLS"}, {"DIR"}, {"DEL"}, {"REN"}, {"VER"},
	{"VOL"},	{"CTTY"}, {"CHCP"}, {"TYPE"}, {"COPY"}, {"DATE"},
	{"TIME"}, {"ERASE"}, {"MKDIR"}, {"RMDIR"}, {"BREAK"}, {"RENAME"},
	{"DELETE"}, {"VERIFY"}, {"TRUENAME"}};

struct {
   char	*szName;
   BYTE	ucKey;
} SKList[] = {
	{"bksp", VK_BACK}, {"break", VK_CANCEL}, {"capslock", VK_CAPITAL},
	{"clear",VK_CLEAR}, {"delete", VK_DELETE}, {"down", VK_DOWN},
	{"end" , VK_END}, {"enter", VK_RETURN}, {"esc", VK_ESCAPE},
	{"help", VK_HELP}, {"home", VK_HOME}, {"insert", VK_INSERT},
	{"left", VK_LEFT}, {"numlock", VK_NUMLOCK}, {"pgdn", VK_NEXT},
	{"pgup", VK_PRIOR}, {"prtsc", VK_SNAPSHOT}, {"pause", VK_PAUSE},
	{"right", VK_RIGHT}, {"tab", VK_TAB}, {"up", VK_UP}, {"scroll", VK_SCROLL},
	{"F1", VK_F1}, {"F2", VK_F2}, {"F3", VK_F3}, {"F4", VK_F4},
	{"F5", VK_F5}, {"F6", VK_F6}, {"F7", VK_F7}, {"F8", VK_F8},
	{"F9", VK_F9}, {"F10", VK_F10}, {"F11", VK_F11}, {"F12", VK_F12},
	{"F13", VK_F13}, {"F14", VK_F14}, {"F15", VK_F15}, {"F16", VK_F16},
	{"F17", VK_F17}, {"F18", VK_F18}, {"F19", VK_F19}, {"F20", VK_F20},
	{"F21", VK_F21}, {"F22", VK_F22}, {"F23", VK_F23}, {"F24", VK_F24},
};		

PTOKEN	ptVarList;
UINT		wVarCount;
PBYTE    pbVarBuff, pbVarStrings;
LPBYTE	lpbTokBuff, lpbTokLine;
HANDLE	hTokBuff, hVarBuff;
INT		sLibCount = 0;
FARPROC	lpfnLibFunc[MAXLIBS];
HINSTANCE hLib[MAXLIBS];
FARPROC	lpfnLibErr[MAXLIBS];
FARPROC	lpfnTokenFuncCallback;
FARPROC	lpfnNotifyRegister = 0;
FARPROC	lpfnNotifyUnRegister;
FARPROC	lpfnTaskFindHandle;
FARPROC	lpfnTHCallBack;
HANDLE	hHoldTask;
HANDLE	hToolHelp;
LPIRQFUNC lpfnIRQ = 0;
UINT		wWinCmdVer = 14;
char		szNull[]= "";
char		szProgName[13] = "";
char		szCmdBuff[128] = "";
char		szProgExt[40] = "";
char		szTemp[33];
BOOL		bReset = FALSE;
BOOL		bRunning = FALSE;
BOOL		bSingleStep = FALSE;
BOOL		bProg =TRUE;
BOOL		bAutoStart = FALSE;
BOOL		bAtEnd = FALSE;
UINT		wCurrentLine;
INT		sFinalRC = 0;
UINT     hWaitTimer = 0;
UINT     hTimer2 = 0;
HANDLE	hDropData;
LONG		fIntFlags;    
//============================================================
// WinMain -- entry point for this application from Windows.
//============================================================
INT APIENTRY WinMain(HANDLE hInstance, HANDLE hPrevInstance, 
                     LPSTR lpCmdLine, INT nCmdShow) {
	INT	rc, sFunc;
	

	hInst = hInstance;
	//
	// If first instance, perform any init processing
	//	
   if(!hPrevInstance)
		if((rc = InitApp(hInstance)) != 0)
			return rc;
	//
	// Initialize this instance
	//
	if((rc = InitInstance(hInstance, lpCmdLine, nCmdShow)) != 0)
		return rc;
	//
	// Application message loop
	//
	bProg = TRUE;
	sFunc = FUNC_EXECUTE;
	while (bProg) {
		//Execute Line has a yield call in it
	   rc = ExecuteLine (&(LPLINETOKEN)lpbTokLine, sFunc);
	   if (rc == ERR_RESET) {
			rc = ResetInterpeter (hMain);
			SetRunning (FALSE);
			bReset = FALSE;
	   }
	   if (rc == ERR_UNEXP_PROGTERM)
	      bProg = FALSE;
	   else if (rc) {
			SetRunning (FALSE);
			bAtEnd = TRUE;
			DragAcceptFiles (hMain, FALSE);
			if (rc == RC_END_OF_PROGRAM) {
				if (bAutoStart)
				   PostMessage (hMain, WM_QUIT, 0, 0);
				else
					PostMessage (hMain, MYMSG_REFRESH, 0, 0);
			} else {
				PrintError (hMain, wCurrentLine, rc, TRUE);
				sFunc = FUNC_EXECUTE | FUNC_NOREFRESH;
			}	
		}
	}	
	//
	// Instance cleanup
	//
	return TermInstance(hInstance, sFinalRC);
}
  
//-----------------------------------------------------------
// InitApp - Global initialization code for this application.
//-----------------------------------------------------------
INT InitApp(HANDLE hInstance) {
	WNDCLASS 	wc;
	//
	// Register App Window class
	//
	wc.style = 0;                             // Window style
	wc.lpfnWndProc = MainWndProc;             // Callback function
	wc.cbClsExtra = 0;                        // Extra class data
	wc.cbWndExtra = 0;                        // Extra window data
	wc.hInstance = hInstance;                 // Owner handle
	wc.hIcon = LoadIcon(hInst, szIconName);   // Application icon
	wc.hCursor = LoadCursor(NULL, IDC_ARROW); // Default cursor
	wc.hbrBackground = COLOR_WINDOW + 1;      // Background color 
	wc.lpszMenuName =  szMenuName;            // Menu name
	wc.lpszClassName = szAppName;             // Window class name

	if (RegisterClass(&wc) == 0)
		return 1;
	
	return 0;
}

//-----------------------------------------------------------
// InitInstance - Instance initialization code for this app.
//-----------------------------------------------------------
INT InitInstance(HANDLE hInstance, LPSTR lpCmdLine, INT nCmdShow) {
   int	i, x, y, Cx, Cy, rc = 0;
   char	szErrStr[128];
	FARPROC	lpfnLibCall;
	HMENU		hMenu;

	x  = GetPrivateProfileInt (szAppName, PRO_XPOS, CW_USEDEFAULT, 
	                           szProfileName);
	y  = GetPrivateProfileInt (szAppName, PRO_YPOS, CW_USEDEFAULT, 
	                           szProfileName);
	Cx = GetPrivateProfileInt (szAppName, PRO_XSIZE, CW_USEDEFAULT, 
	                           szProfileName);
	Cy = GetPrivateProfileInt (szAppName, PRO_YSIZE, CW_USEDEFAULT, 
	                           szProfileName);

	GetProfileString ("Windows", "Programs", "com exe pif bat",
	                  szProgExt, sizeof (szProgExt));
	hMain = CreateWindow(                     // create frame window
	            szAppName,                    // window class name
	            szTitleText,                  // text for title bar
	            WS_OVERLAPPEDWINDOW,          // window style
	            x, y,                         // default position
	            Cx, Cy,                       // default size
	            NULL,                         // no parent window
	            NULL,                         // use class default menu
	            hInstance,                    // window owner
	            NULL);                        // ptr to creation data

	if(!hMain) return 0x10;
	lpfnTokenFuncCallback = MakeProcInstance ((FARPROC) TokFuncCallback, hInstance);	
	rc = ParseCmdLine (lpCmdLine);
	if (rc) {
		if (rc > 0)
		   rc += 11;
		else 
		   rc = abs (rc);
		if (LoadString (hInst, rc, szErrStr, sizeof (szErrStr)) == 0) {
			itoa (rc, szTemp, 10);
			strcpy (szErrStr, "Error number: ");
			strcat (szErrStr, szTemp);
		}
		MessageBox (hMain, szErrStr, szAppName, MB_OK | MB_ICONHAND);
		return 0;
	}
	rc = InitInterpeter (hMain);
	if (rc) {
		if (!GetErrorString (-1, rc, szErrStr, sizeof (szErrStr)))
			strcpy (szErrStr, "Error initializing interpeter");
		MessageBox (hMain, szErrStr, szAppName, MB_OK | MB_ICONHAND);
		return rc;
	}
	// Add Lib About items to help menu
	hMenu = GetMenu (hMain);
	for (i = 0; i < sLibCount; i++) {
		lpfnLibCall = GetProcAddress (hLib[i], "WCGetVersion");
		strcpy (szErrStr, "About ");
		((LPLIBGETVER)*lpfnLibCall) ((LPINT) &x, 
		                           (LPSTR) szErrStr + strlen(szErrStr), 
		                           sizeof (szErrStr) - strlen (szErrStr)-3);
		strcat (szErrStr, "...");
		AppendMenu (GetSubMenu (hMenu, GetMenuItemCount(hMenu)-1), MF_ENABLED,
		            IDM_ABOUT+i+1, szErrStr);
	}			   
	if (bAutoStart) 
	   ShowWindow(hMain, nCmdShow | SW_SHOW | SW_MINIMIZE);
	else
		ShowWindow(hMain, nCmdShow | SW_SHOW);
	UpdateWindow(hMain);              // force WM_PAINT message
	return 0;                         // return success flag
}

//------------------------------------------------------------
// TermInstance - Instance termination code for this app.
//------------------------------------------------------------
INT TermInstance(HANDLE hinstance, int sDefRC) {
   TermInterpeter ();
	return sDefRC;
}

//============================================================
// Message handling procedures for MainWindow
//============================================================
//------------------------------------------------------------
// OutputDebugLong - Write a number to the debug console
//------------------------------------------------------------
void OutputDebugLong (LONG lNum) {
   char szStr[34];

	ltoa (lNum, szStr, 16);
	OutputDebugString (szStr);
	return;
}				               
//------------------------------------------------------------
// MainWndProc - Callback function for application window
//------------------------------------------------------------
LONG CALLBACK MainWndProc(HWND hWnd, UINT wMsg, UINT wParam, 
                          LONG lParam) {
	INT i;
	//
	// Search message list to see if we need to handle this
	// message.  If in list, call procedure.
	//
	for(i = 0; i < dim(MainMessages); i++) {
		if(wMsg == MainMessages[i].Code)
			return (*MainMessages[i].Fxn)(hWnd, wMsg, wParam, lParam);
	}
	return DefWindowProc(hWnd, wMsg, wParam, lParam);
}

//------------------------------------------------------------
// GetLListPos - Computes the position of the source listbox
//------------------------------------------------------------ 
void GetLListPos (HWND hWnd, INT *x, INT *y, INT *Cx, INT *Cy, UINT *fFlags){
   RECT	rect;
   INT	sWinHeight, sWinWidth;

	GetClientRect (hWnd, &rect);
	sWinHeight = rect.bottom - rect.top;
	sWinWidth = rect.right - rect.left;

   if (bShowList) 
		*fFlags = SWP_SHOWWINDOW;
	else	
		*fFlags = SWP_HIDEWINDOW;

  	*x = 10;
  	*Cx = sWinWidth - 90;
   if (bShowOut) {
   	*y = sWinHeight/2 + 10;
   	*Cy = sWinHeight/2 - 10;
   } else {
   	*y = 10;
   	*Cy = sWinHeight - 10;
   }
	return;
}	
//------------------------------------------------------------
// GetOListPos - Computes the position of the source listbox
//------------------------------------------------------------ 
void GetOListPos (HWND hWnd, INT *x, INT *y, INT *Cx, INT *Cy, UINT *fFlags){
   RECT	rect;
   INT	sWinHeight, sWinWidth;

	GetClientRect (hWnd, &rect);
	sWinHeight = rect.bottom - rect.top;
	sWinWidth = rect.right - rect.left;

   if (bShowOut) 
		*fFlags = SWP_SHOWWINDOW;
	else	
		*fFlags = SWP_HIDEWINDOW;

  	*x = 10;
  	*y = 10;
  	*Cx = sWinWidth - 90;
   if (bShowList) 
   	*Cy = sWinHeight/2 - 10;
   else
   	*Cy = sWinHeight - 10;
	return;
}	

//------------------------------------------------------------
// GetBtnPos - Computes the position of the buttons
//------------------------------------------------------------ 
void GetBtnPos (HWND hWnd, INT *x, INT *y, INT *Cx, INT *Cy, INT sNum) {
   RECT	rect;
   INT	sWinHeight, sWinWidth;

	GetClientRect (hWnd, &rect);
	sWinHeight = rect.bottom - rect.top;
	sWinWidth = rect.right - rect.left;

	*x = sWinWidth - 75;
	*y = 30 + (35 * sNum);
  	*Cx = 70;
  	*Cy = 30;
	return;
}	
//------------------------------------------------------------
// DoCreateMain - process WM_CREATE message for frame window.
//------------------------------------------------------------ 
LONG DoCreateMain (HWND hWnd, UINT wMsg, UINT wParam, LONG lParam) {
   DWORD	dwStyle;
   INT	x,y,Cx,Cy;
   UINT	fFlags;
   RECT	rect;
   INT	sWinWidth;

	GetClientRect (hWnd, &rect);
	sWinWidth = rect.right - rect.left;

	//Create Child Windows 
	GetLListPos (hWnd, &x, &y, &Cx, &Cy, &fFlags);
   if (bShowList) 
		dwStyle = WS_CHILD | WS_VISIBLE | WS_BORDER | WS_VSCROLL |
		          LBS_NOTIFY | LBS_DISABLENOSCROLL | LBS_USETABSTOPS;
	else
		dwStyle = WS_CHILD | WS_BORDER | WS_VSCROLL |
		          LBS_NOTIFY | LBS_DISABLENOSCROLL | LBS_USETABSTOPS;
	hwndLList = CreateWindow ("listbox", NULL, dwStyle,
                     x, y,                             // Pos
                     Cx, Cy,                           // Size
                     hWnd,                             //Parent handle
                     IDD_LINELIST,                     //ID number
                     hInst, NULL);

	GetOListPos (hWnd, &x, &y, &Cx, &Cy, &fFlags);	
   if (bShowOut) 
		dwStyle = WS_CHILD | WS_VISIBLE | WS_BORDER | WS_VSCROLL |
		          LBS_NOTIFY | LBS_DISABLENOSCROLL | LBS_USETABSTOPS;
	else	
		dwStyle = WS_CHILD | WS_BORDER | WS_VSCROLL |
		          LBS_NOTIFY | LBS_DISABLENOSCROLL;
	hwndOList = CreateWindow ("listbox", NULL, dwStyle,
                     x, y,                             // Pos
                     Cx, Cy,                           // Size
                     hWnd,                             //Parent handle
                     IDD_OUTLIST,                      //ID number
                     hInst, NULL);

	x = 14;
	SendDlgItemMessage (hWnd, IDD_LINELIST, LB_SETTABSTOPS, 1, 
	                    (LPARAM)(LPINT)&x);
	SendDlgItemMessage (hWnd, IDD_OUTLIST, LB_SETTABSTOPS, 1, 
	                    (LPARAM)(LPINT)&x);

	GetBtnPos (hWnd, &x, &y, &Cx, &Cy, 0);
	hwndRunBtn = CreateWindow ("button", "Run",
                     WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                     x, y,                             // Pos
                     Cx, Cy,                           // Size
                     hWnd,                             //Parent handle
                     IDD_RUN,                          //ID number
                     hInst, NULL);
	GetBtnPos (hWnd, &x, &y, &Cx, &Cy, 1);
	hwndStepBtn = CreateWindow ("button", "Step",
                     WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                     x, y,                             // Pos
                     Cx, Cy,                           // Size
                     hWnd,                             //Parent handle
                     IDD_STEP,                         //ID number
                     hInst, NULL);
	GetBtnPos (hWnd, &x, &y, &Cx, &Cy, 2);
	hwndResetBtn = CreateWindow ("button", "Reset",
                     WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                     x, y,                             // Pos
                     Cx, Cy,                           // Size
                     hWnd,                             //Parent handle
                     IDD_RESET,                        //ID number
                     hInst, NULL);

	return 0;         
}
//------------------------------------------------------------
// DoSizeMain - process WM_SIZE message for frame window.
//------------------------------------------------------------ 
LONG DoSizeMain (HWND hWnd, UINT wMsg, UINT wParam, LONG lParam) {

   SendMessage (hWnd, MYMSG_REFRESH, 0, 0);
	return 0;         
}
//------------------------------------------------------------
// DoTimerMain - process WM_TIMER message for frame window.
//------------------------------------------------------------ 
LONG DoTimerMain (HWND hWnd, UINT wMsg, UINT wParam, LONG lParam) {

	if (wParam == WAITTIMER_ID)
		SetRunning (TRUE);
	return 0;         
}
//------------------------------------------------------------
// DoDropFilesMain - process WM_DROPFILES message for frame window.
//------------------------------------------------------------ 
LONG DoDropFilesMain (HWND hWnd, UINT wMsg, UINT wParam, LONG lParam) {

	DropFiles(hWnd, (HANDLE)wParam);
	return 0;         
}
//------------------------------------------------------------
// DoInitMenuMain - process WM_INITMENU message
//------------------------------------------------------------ 
LONG DoInitMenuMain (HWND hWnd, UINT wMsg, UINT wParam, LONG lParam) {

	if (bShowList)	
		CheckMenuItem (GetMenu(hWnd), IDM_SHOWLIST,
		               MF_BYCOMMAND | MF_CHECKED);
	else		               
		CheckMenuItem (GetMenu(hWnd), IDM_SHOWLIST,
		               MF_BYCOMMAND | MF_UNCHECKED);
	if (bShowOut)	
		CheckMenuItem (GetMenu(hWnd), IDM_SHOWOUT,
		               MF_BYCOMMAND | MF_CHECKED);
	else		               
		CheckMenuItem (GetMenu(hWnd), IDM_SHOWOUT,
		               MF_BYCOMMAND | MF_UNCHECKED);
	if (bMinOnRun)	
		CheckMenuItem (GetMenu(hWnd), IDM_MINONRUN,
		               MF_BYCOMMAND | MF_CHECKED);
	else		               
		CheckMenuItem (GetMenu(hWnd), IDM_MINONRUN,
		               MF_BYCOMMAND | MF_UNCHECKED);
	return 0;         
}
//------------------------------------------------------------
// DoRefreshMain - Refresh main windows.  This procedure
// is called by the user defined MYMSG_REFRESH message
//------------------------------------------------------------ 
LONG DoRefreshMain (HWND hWnd, UINT wMsg, UINT wParam, LONG lParam) {
   INT	x,y,Cx,Cy;
   UINT	fFlags;

	GetLListPos (hWnd, &x, &y, &Cx, &Cy, &fFlags);
	SetWindowPos (hwndLList, 0, x, y, Cx, Cy, fFlags);
	GetOListPos (hWnd, &x, &y, &Cx, &Cy, &fFlags);
	SetWindowPos (hwndOList, 0, x, y, Cx, Cy, fFlags);

	GetBtnPos (hWnd, &x, &y, &Cx, &Cy, 0);	
	SetWindowPos (hwndRunBtn, 0, x, y, Cx, Cy, SWP_NOZORDER);
	GetBtnPos (hWnd, &x, &y, &Cx, &Cy, 1);	
	SetWindowPos (hwndStepBtn, 0, x, y, Cx, Cy, SWP_NOZORDER);
	GetBtnPos (hWnd, &x, &y, &Cx, &Cy, 2);	
	SetWindowPos (hwndResetBtn, 0, x, y, Cx, Cy, SWP_NOZORDER);

	if (szCmdBuff[0] == 0) {
		EnableWindow (GetDlgItem (hWnd, IDD_STEP), FALSE);
		EnableWindow (GetDlgItem (hWnd, IDD_RESET), FALSE);
		SetDlgItemText (hWnd, IDD_RUN, "Run");
		EnableWindow (GetDlgItem (hWnd, IDD_RUN), FALSE);
	} else {
		EnableWindow (GetDlgItem (hWnd, IDD_RUN), TRUE);
		if (bRunning) {
			EnableWindow (GetDlgItem (hWnd, IDD_STEP), FALSE);
			EnableWindow (GetDlgItem (hWnd, IDD_RESET), FALSE);
			SetDlgItemText (hWnd, IDD_RUN, "Stop");
		} else {
			SetDlgItemText (hWnd, IDD_RUN, "Run");
			EnableWindow (GetDlgItem (hWnd, IDD_RESET), TRUE);
		   if (bAtEnd) {
				EnableWindow (GetDlgItem (hWnd, IDD_STEP), FALSE);
				EnableWindow (GetDlgItem (hWnd, IDD_RUN), FALSE);
			} else
				EnableWindow (GetDlgItem (hWnd, IDD_STEP), TRUE);
		}	
	}
	SendDlgItemMessage (hMain, IDD_LINELIST, LB_SETCURSEL, wCurrentLine, 0);
	return 0;         
}
//------------------------------------------------------------
// DoRestartMain - Process MYMSG_RESTART message.  Restart 
// program. 
//------------------------------------------------------------ 
LONG DoRestartMain (HWND hWnd, UINT wMsg, UINT wParam, LONG lParam) {

	SetRunning (wParam);
	return 0;
}
//------------------------------------------------------------
// DoCommandMain - process WM_COMMAND message for frame window 
// by decoding the menubar item with the menuitems[] array, 
// then running the corresponding function to process the command.
//------------------------------------------------------------ 
LONG DoCommandMain (HWND hWnd, UINT wMsg, UINT wParam, LONG lParam) {
	INT	i;
	UINT	idItem, wNotifyCode, wVer;
	HWND	hwndCtl;
	FARPROC	lpfnLibCall;

	idItem = (UINT) wParam;                      // Parse Parameters
	hwndCtl = (HWND) LOWORD(lParam);
	wNotifyCode = (UINT) HIWORD(lParam);
	//
	// Call routine to handle control message
	//
	for(i = 0; i < dim(MainMenuItems); i++) {
		if(idItem == MainMenuItems[i].Code)
			return (*MainMenuItems[i].Fxn)(hWnd, idItem, hwndCtl, 
			                               wNotifyCode);
	}
	//
	// Special check for About menu items
	//
	if ((idItem > IDM_ABOUT) && (idItem < IDM_ABOUT+1+MAXLIBS)) {
		i = idItem - IDM_ABOUT - 1;
		lpfnLibCall = GetProcAddress (hLib[i], "WCGetVersion");
		((LPLIBGETVER)*lpfnLibCall) ((LPWORD) &wVer, (LPSTR)-1, 0);
		if (wVer)
			MyDisplayDialog(hLib[i], "AboutBox", hWnd, 
   	                 (WNDPROC) AboutDlgProc, (LONG) wVer);
	}
	return DefWindowProc(hWnd, wMsg, wParam, lParam);
}

//------------------------------------------------------------
// DoDestroyMain - process WM_DESTROY message for frame window.
//------------------------------------------------------------ 
LONG DoDestroyMain (HWND hWnd, UINT wMsg, UINT wParam, LONG lParam) {
   RECT	rect;

   bReset = TRUE;
   bAtEnd = TRUE;
	//Save Window position
	if	(!IsIconic (hWnd) && !IsZoomed (hWnd)) {
		GetWindowRect (hWnd, &rect);
		MyWritePrivateProfileInt (szAppName, PRO_XPOS, rect.left,
		                          10, szProfileName);
		MyWritePrivateProfileInt (szAppName, PRO_YPOS, rect.top,
		                          10, szProfileName);
		MyWritePrivateProfileInt (szAppName, PRO_XSIZE, 
		                          rect.right - rect.left,
		                          10, szProfileName);
		MyWritePrivateProfileInt (szAppName, PRO_YSIZE, 
		                          rect.bottom - rect.top,
		                          10, szProfileName);
	}
	PostQuitMessage (0);
	return 0;         
}

//============================================================
// Control handling procedures for MainWindow
//============================================================
//------------------------------------------------------------
// DoMainCtlRun - Process Run button
//------------------------------------------------------------ 
LONG DoMainCtlRun (HWND hWnd, UINT idItem, HWND hwndCtl, 
                     UINT wNotifyCode) {
	if (bRunning) {
		SetRunning (FALSE);
		bShowList = TRUE;

	} else {	
		SetRunning (TRUE);
		bSingleStep = FALSE;
		bShowList = FALSE;
		bShowOut = TRUE;
		if (bMinOnRun)
			   ShowWindow (hWnd, SW_SHOWMINIMIZED);
		else
		   return 0;
	}
	return 0;
}
//------------------------------------------------------------
// DoMainCtlStep - Process Step button
//------------------------------------------------------------ 
LONG DoMainCtlStep (HWND hWnd, UINT idItem, HWND hwndCtl, 
                     UINT wNotifyCode) {

	bSingleStep = TRUE;
	SetRunning (TRUE);
	bShowList = TRUE;
	bShowOut = TRUE;
	PostMessage (hWnd, MYMSG_REFRESH, 0, 0);
	return 0;
}
//------------------------------------------------------------
// DoMainCtlReset - Process Reset button
//------------------------------------------------------------ 
LONG DoMainCtlReset (HWND hWnd, UINT idItem, HWND hwndCtl, 
                     UINT wNotifyCode) {
	bReset = TRUE;
	bAtEnd = FALSE;
	SetRunning (TRUE);
	bShowList = TRUE;
	bShowOut = FALSE;
	return 0;
}

//============================================================
// Menu handling procedures for MainWindow
//============================================================
//------------------------------------------------------------
// DoMainMenuFileOpen - Process File/Open menu item
//------------------------------------------------------------ 
LONG DoMainMenuFileOpen (HWND hWnd, UINT idItem, HWND hwndCtl, 
                         UINT wNotifyCode) {
	static char *szFilter[] = {"WinCmd Files (*.WCM)", "*.WCM",
	                           "All Files (*.*)", "*.*", "" };

	if (MyGetFilename (hWnd, szCmdBuff, sizeof (szCmdBuff)-1, *szFilter, 1) == 0)
		return 0;
	szCmdBuff[strlen (szCmdBuff)+1] = 0;

	bReset = TRUE;
	SetRunning (TRUE);
	bShowList = TRUE;
	bShowOut = FALSE;
	return 0;
}
//------------------------------------------------------------
// DoMainMenuShowList - Process View/Show Listing menu item
//------------------------------------------------------------ 
LONG DoMainMenuShowList (HWND hWnd, UINT idItem, HWND hwndCtl, 
                         UINT wNotifyCode) {

	if (bShowList)
	   bShowList = FALSE;
	else
	   bShowList = TRUE;
	SendMessage (hWnd, MYMSG_REFRESH, 0, 0);
	return 0;
}
//------------------------------------------------------------
// DoMainMenuShowOut - Process View/Show Output menu item
//------------------------------------------------------------ 
LONG DoMainMenuShowOut (HWND hWnd, UINT idItem, HWND hwndCtl, 
                         UINT wNotifyCode) {

	if (bShowOut)
	   bShowOut = FALSE;
	else
	   bShowOut = TRUE;
	SendMessage (hWnd, MYMSG_REFRESH, 0, 0);
	return 0;
}
//------------------------------------------------------------
// DoMainMenuClearOut - Process View/Clear Output menu item
//------------------------------------------------------------ 
LONG DoMainMenuClearOut (HWND hWnd, UINT idItem, HWND hwndCtl, 
                         UINT wNotifyCode) {

	SendDlgItemMessage (hWnd, IDD_OUTLIST, LB_RESETCONTENT, 0, 0);
	return 0;
}
//------------------------------------------------------------
// DoMainMenuMinOnRun - Process View/Minimize on Run menu item
//------------------------------------------------------------ 
LONG DoMainMenuMinOnRun (HWND hWnd, UINT idItem, HWND hwndCtl, 
                         UINT wNotifyCode) {

	if (bMinOnRun)
	   bMinOnRun = FALSE;
	else
	   bMinOnRun = TRUE;
	return 0;
}
//------------------------------------------------------------
// DoMainMenuFileExit - Process File/Exit menu item
//------------------------------------------------------------ 
LONG DoMainMenuFileExit (HWND hWnd, UINT idItem, HWND hwndCtl, 
                         UINT wNotifyCode) {

	DestroyWindow (hWnd);  //Destroy window
	return 0;
}
//------------------------------------------------------------
// DoMenuHelp - Process Help-Help command from menu bar.
//------------------------------------------------------------ 
LONG DoMainMenuHelp (HWND hWnd, UINT idItem, HWND hwndCtl, 
                      UINT wNotifyCode) {

	MyDisplayDialog(hInst, "HelpBox", hWnd, 
   	             (WNDPROC) HelpDlgProc, 0);
	return 0;
}
//------------------------------------------------------------
// DoMenuAbout - Process Help-About command from menu bar.
//------------------------------------------------------------ 
LONG DoMainMenuAbout (HWND hWnd, UINT idItem, HWND hwndCtl, 
                      UINT wNotifyCode) {

	MyDisplayDialog(hInst, "AboutBox", hWnd, 
   	             (WNDPROC) AboutDlgProc, (LONG) wWinCmdVer);
	return 0;
}
//============================================================
// HelpDlgProc - Help dialog box dialog procedure
//============================================================
BOOL CALLBACK HelpDlgProc (HWND hWnd, UINT msg, UINT wParam, 
                            LONG lParam) {
	char		szHelpStr[128];
	FARPROC	lpfnLibCall;
	INT		i, j;
                              
	if (msg == WM_INITDIALOG) {
		SendDlgItemMessage (hWnd, IDD_HELPLIST, LB_RESETCONTENT, 0, 0);
		i = 0;			
		while (LoadString (hInst, WCH+i, szHelpStr, sizeof (szHelpStr)) != 0) {
			SendDlgItemMessage (hWnd, IDD_HELPLIST, LB_ADDSTRING, 0, 
			                    (LPARAM) (LPSTR) szHelpStr);
			i++;			                    
		}
		SendDlgItemMessage (hWnd, IDD_HELPLIST, LB_SETHORIZONTALEXTENT, 
   		                 1000, 0L);

		// Add Library help lines
	   for (j = 0; j < sLibCount; j++) {
			lpfnLibCall = GetProcAddress (hLib[j], "WCGetHelp");
			if (lpfnLibCall != 0) {
			   szHelpStr[0] = '\0';
				SendDlgItemMessage (hWnd, IDD_HELPLIST, LB_ADDSTRING, 0, 
				                    (LPARAM) (LPSTR) szHelpStr);
				SendDlgItemMessage (hWnd, IDD_HELPLIST, LB_ADDSTRING, 0, 
				                    (LPARAM) (LPSTR) szHelpStr);
				i = 0;			
				while (((LPLIBGETHELP)*lpfnLibCall) (i, (LPSTR) szHelpStr, 
				                                     sizeof (szHelpStr))) {
					SendDlgItemMessage (hWnd, IDD_HELPLIST, LB_ADDSTRING, 0, 
					                    (LPARAM) (LPSTR) szHelpStr);
					i++;			                    
				}
			}
		}	
	   szHelpStr[0] = '\0';
		SendDlgItemMessage (hWnd, IDD_HELPLIST, LB_ADDSTRING, 0, 
		                    (LPARAM) (LPSTR) szHelpStr);
		SetFocus (GetDlgItem (hWnd, IDD_HELPLIST));   		                 
	}   
	if((msg == WM_COMMAND) && ((wParam == IDOK) || (wParam == IDCANCEL))) {
		EndDialog(hWnd, 0);
		return TRUE;
	} 
	return FALSE;
}
//============================================================
// AboutDlgProc - About dialog box dialog procedure
//============================================================
BOOL CALLBACK AboutDlgProc (HWND hWnd, UINT msg, UINT wParam, 
                            LONG lParam) {
	char	szAboutStr[128];
	
	HWND	hwndText;
	RECT	rect;
	HDC	hdc;
	PAINTSTRUCT ps;
	HPEN	hDPen, hLPen, hOldPen;

	switch (msg) {                              
		case WM_INITDIALOG:
			GetDlgItemText (hWnd, IDD_PROGSTR, szAboutStr, sizeof (szAboutStr));
			itoa ((INT)lParam/10, &szAboutStr[strlen (szAboutStr)], 10);
			strcat (szAboutStr, ".");
			itoa ((INT)lParam%10, &szAboutStr[strlen (szAboutStr)], 10);
			SetDlgItemText (hWnd, IDD_PROGSTR, szAboutStr);
			return TRUE;
	
		case WM_PAINT:
			hdc = BeginPaint(hWnd, &ps);
			hwndText = GetDlgItem (hWnd, IDD_COPYDATE);
			if (IsWindow (hwndText)) {
				GetClientRect (hwndText, &rect);
				ClientToScreen (hwndText, (LPPOINT)&rect);
				ScreenToClient (hWnd, (LPPOINT)&rect);
				rect.left -= 2;
				rect.top -= 2;
				rect.right += rect.left + 4;
				rect.bottom += rect.top + 4;
				hDPen = CreatePen (PS_SOLID, 1, GetSysColor (COLOR_BTNSHADOW));
				hLPen = CreatePen (PS_SOLID, 1, GetSysColor (COLOR_BTNHIGHLIGHT));

				hOldPen = SelectObject (hdc, hDPen);
				MoveTo (hdc, rect.left, rect.bottom);
				LineTo (hdc, rect.left, rect.top);
				LineTo (hdc, rect.right+1, rect.top);
				//Start at bottom left, draw light pen over and up.
				SelectObject (hdc, hLPen);
				MoveTo (hdc, rect.left+1, rect.bottom);
				LineTo (hdc, rect.right, rect.bottom);
				LineTo (hdc, rect.right, rect.top);
				SelectObject (hdc, hOldPen);

				DeleteObject (hDPen);
				DeleteObject (hLPen);
			}	
			EndPaint(hWnd, &ps);
			return TRUE;
		
		case WM_COMMAND:
		   if ((wParam == IDOK) || (wParam == IDCANCEL)) {
				EndDialog(hWnd, 0);
				return TRUE;
			} 
	} 
	return FALSE;
}
//============================================================
// AskDlgProc - Ask box dialog box dialog procedure
//============================================================
BOOL CALLBACK AskDlgProc (HWND hWnd, UINT msg, UINT wParam, 
                            LONG lParam) {
	RECT	rectDesk, rectWnd;                              
	INT	x, y;

	if (msg == WM_INITDIALOG) {
	   //First, center the box on the screen.
	   GetWindowRect (GetDesktopWindow(), &rectDesk);
	   GetWindowRect (hWnd, &rectWnd);
	   x = (rectDesk.right-rectDesk.left)/2-(rectWnd.right-rectWnd.left)/2;
	   y = (rectDesk.bottom-rectDesk.top)/2-(rectWnd.bottom-rectWnd.top)/2;
	   SetWindowPos (hWnd, 0, x, y, 0, 0, SWP_NOSIZE);

	   //Set the approperate text fields.
		SetWindowText (hWnd, szProgName);
	  	SetDlgItemText (hWnd, IDD_ASKTEXT, (char *)LOWORD (lParam));
   	SetDlgItemText (hWnd, IDD_ASKANS, (char *)HIWORD (lParam));
		return 1;	   	                
	}	   	
                              
	if (msg == WM_COMMAND) {
		if(wParam == IDOK) {
		   GetDlgItemText (hWnd, IDD_ASKANS, pszAskAns, MAXLINELEN);
			EndDialog(hWnd, 1);
			return TRUE;
		} else if (wParam == IDCANCEL) {
			EndDialog(hWnd, 0);
			return TRUE;
		}	
	} 
	return FALSE;
}

//============================================================  
// General Helper Routines 
//============================================================  
//------------------------------------------------------------
// MyYield - Yields control to other programs, but returns
// if Windows is idle.
//------------------------------------------------------------
BOOL MyYield (void) {
   MSG	msg;
   BOOL	bCont;
   
	if (!bRunning) {
		//This loop is much better for power sensitive machines
		while (!bRunning && (bCont = GetMessage (&msg, NULL, 0, 0))) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}	
	} else {
	   bCont = TRUE;
		while (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE)) {
		   if (msg.message == WM_QUIT)
		      bCont = FALSE;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}	
	}
	return bCont;
}
//------------------------------------------------------------
// MyDisplayDialog - Display a dialog box
//------------------------------------------------------------ 
INT MyDisplayDialog (HINSTANCE hInstance, LPCSTR szDlgName,
                     HWND hWnd, WNDPROC lpDialogProc, 
                     LPARAM lParam) {
    WNDPROC lpDlgProcInst;
    INT		rc;

    lpDlgProcInst = MakeProcInstance(lpDialogProc, hInst);
    rc = DialogBoxParam (hInstance, szDlgName, hWnd, 
                         lpDlgProcInst, lParam);
    FreeProcInstance(lpDlgProcInst);
    return rc;                              
}
//------------------------------------------------------------
// MyWritePrivateProfileInt - Writes an integer to the profile
//------------------------------------------------------------
BOOL MyWritePrivateProfileInt (char *szSec, char *szEntry, 
                               int Num, int Base, char *szProfile) {
	char	szStr[33];
	                           
	itoa (Num, szStr, Base);
	return WritePrivateProfileString (szSec, szEntry, szStr, 
	                                  szProfile);
}

//------------------------------------------------------------
// MyGetFilename - Uses the common File Open dialog box to
// query the user for a filename
//------------------------------------------------------------
BOOL MyGetFilename (HWND hWnd, char *szFilename, INT sMaxLen, char *szFilter,
                    INT sFilterIndex) {
   OPENFILENAME ofn;

   *szFilename = '\0';
	memset (&ofn, 0, sizeof (ofn));

	ofn.lStructSize = sizeof (OPENFILENAME);
	ofn.hwndOwner = hWnd;
	ofn.hInstance = hInst;
	ofn.lpstrFilter = szFilter;
	ofn.nFilterIndex = sFilterIndex;
	ofn.lpstrFile = szFilename;
	ofn.nMaxFile = sMaxLen;
	ofn.Flags = OFN_FILEMUSTEXIST;
	ofn.lpstrDefExt = "WCM";

 	return (GetOpenFileName (&ofn));
}


//------------------------------------------------------------
// MySubClassWindow - Subclasses a window 
//------------------------------------------------------------
WNDPROC MySubClassWindow (HWND hWnd, WNDPROC lpfnNewProc) {
   WNDPROC lpfnOldProc;

	lpfnOldProc = (WNDPROC) GetWindowLong (hWnd, GWL_WNDPROC);
	SetWindowLong (hWnd, GWL_WNDPROC, (LONG) lpfnNewProc);
	return lpfnOldProc;				               
}				               

//============================================================  
// Program specific procedures
//============================================================
//============================================================
// Start of Interpeter code
//============================================================
//------------------------------------------------------------
// InitInterpeter - Initializes the interpeter 
//------------------------------------------------------------ 
INT InitInterpeter (HWND hWnd) {
   INT		sRC;
	FARPROC	lpfnLibCall;
	FIND_T	fs;
	char		szDir[260];

	hVarBuff = LocalAlloc (LMEM_MOVEABLE, (UINT) VARBUFFSIZE);
	if (hVarBuff == 0) return ERR_OUT_OF_MEMORY;
	pbVarBuff = LocalLock (hVarBuff);

	hTokBuff = GlobalAlloc (GMEM_MOVEABLE | LMEM_ZEROINIT, TOKBUFFSIZE);
	if (hTokBuff == 0) return ERR_OUT_OF_MEMORY;
	lpbTokBuff = GlobalLock (hTokBuff);

	GetModuleFileName (hInst, szDir, sizeof (szDir));
	strcpy (strrchr (szDir, '\\')+1, "*.WCL");
	sRC = _dos_findfirst (szDir, _A_NORMAL, &fs);
	while (sRC == 0) {
		hLib[sLibCount] = LoadLibrary (fs.name);
		if (hLib[sLibCount] > 32) {
			lpfnLibCall = GetProcAddress (hLib[sLibCount], "WCLibLoad");
			if (lpfnLibCall != 0) {
				sRC = ((LPLIBLOAD)*lpfnLibCall) (hWnd, MAXLIBERRS,
		                          ERR_LIBERROR_START - (sLibCount * MAXLIBERRS));
				if (sRC == 0) {										  
					// Get and save address for calling Library functions
					lpfnLibFunc[sLibCount] = GetProcAddress (hLib[sLibCount], 
					                                         "WCLibFunc");
					// Added in 1.2, allows libs to provide their own error strings
					lpfnLibErr[sLibCount] = GetProcAddress (hLib[sLibCount], 
					                                        "WCGetErrStr");
				} else
				   sRC = 1;
			} else
			   sRC = 1;
		} else
		   sRC = 1;
		if ((sRC) || (lpfnLibFunc[sLibCount] == 0) || (lpfnLibErr[sLibCount] == 0)) {
	   	FreeLibrary (hLib[sLibCount]);
		   sRC = 0;
		} else
			sLibCount++;
		sRC = _dos_findnext (&fs);
	}
	//If loop term due to file not found, RC = 0;
	if ((sRC == 2) || (sRC == 3) || (sRC == 0x12)) sRC = 0;
	if (sRC) return sRC;
	sRC = ResetInterpeter (hWnd);	
	if (sRC) return sRC;
	if (bAutoStart) {
		SetRunning (TRUE);
		bShowList = FALSE;
		bShowOut = TRUE;
	}
	bReset = FALSE;
	return sRC;
}   

//------------------------------------------------------------
// ResetInterpeter - Resets the interpeter and loads a file.
//------------------------------------------------------------ 
INT ResetInterpeter (HWND hWnd) {
   INT		sRC = 0, i, j;
   PTOKEN	ptArg0;
   HCURSOR	hcurSave;
   FARPROC	lpfnLibReset;
	
	pbVarStrings = pbVarBuff;
	ptVarList = (PTOKEN) (pbVarStrings + VARBUFFSIZE);
  	wVarCount = 0;
	lpbTokLine = lpbTokBuff;
	hHoldTask = 0;
	fIntFlags = 0;

	// Add Operators to token list
   for (i = 0; i < dim (OpList); i++) {
		sRC = AddTokenInit (OpList[i].szName, OpList[i].ucType, 
		                    (WORD) OpList[i].ucLevel,
		                    (DWORD) (LPVOID) OpList[i].Fxn);
		if (sRC) break;
	}	
	// Add DOS internal commands to token list
   for (i = 0; i < dim (CmdList); i++) {
		sRC = AddTokenInit (CmdList[i].szName, TTYPE_INTCMD, 0, 0);
		if (sRC) break;
	}	
	// Add Functions to token list
   for (i = 0; i < dim (FuncList); i++) {
		sRC = AddTokenInit (FuncList[i].szName, FuncList[i].ucType, 0,
		                    (DWORD)(LPVOID) FuncList[i].Fxn);
		if (sRC) break;
	}	
	// Add Statements to token list
   for (i = 0; i < dim (StatementList); i++) {
		sRC = AddTokenInit (StatementList[i].szName, 
		                    StatementList[i].ucType, 0,
		                    (DWORD)(LPVOID) StatementList[i].Fxn);
		if (sRC) break;
	}	
	// Load Predefined Vars
	ConstList[0].lData = hInst;
	ConstList[1].lData = hMain;
	ConstList[2].lData = wWinCmdVer;
   for (i = 0; i < dim (ConstList); i++) {
		sRC = AddTokenInit (ConstList[i].szName, TTYPE_NUM, 0,
		                    (DWORD)ConstList[i].lData);
		if (sRC) break;
	}	
	// Load Cmd line arg Tokens
	i = 0;
	ptArg0 = AddToken ("", TTYPE_STR, 0, (DWORD)(LPSTR) szCmdBuff, &sRC);
	if (szCmdBuff[0] != 0) {
		for (i = 0; szCmdBuff[i] != 0; i++)
		   ;
		i++;
	}	
	j = 1;
	while ((szCmdBuff[i] != 0) && (sRC == 0)) {
		AddToken ("",TTYPE_STR, 0, (DWORD)(LPSTR)&szCmdBuff[i], &sRC);
		for (; szCmdBuff[i] != 0; i++)
		   ;
		i++;
		j++;
	}
	if (sRC) return sRC;
	AddToken ("",TTYPE_RETVAL, j, (DWORD)(LPVOID)ptArg0, &sRC);

	// Add Library Tokens
   for (i = 0; i < sLibCount; i++) {
		lpfnLibReset = GetProcAddress (hLib[i], "WCLibReset");
		if (lpfnLibReset == 0)
		   sRC = ERR_BAD_LIB_LOAD;
		else
			sRC = ((LPLIBRESET)*lpfnLibReset) (i, lpfnTokenFuncCallback);
		if (sRC) break;
	}	
	if (sRC) {
	   PrintError (hWnd, -1, sRC, TRUE);
	   return 0;
	}
	// Load file
	hcurSave = SetCursor (LoadCursor (NULL, IDC_WAIT));
	sRC = LoadProgram (hWnd, (LPLINETOKEN) lpbTokLine, (UINT)TOKBUFFSIZE);
	SetCursor (hcurSave);
	bSingleStep = FALSE;
	bAtEnd = FALSE;
	wCurrentLine = 0;
	if (hTimer2) {
	   KillTimer (hMain, hTimer2);
		hTimer2 = 0;
	}	
	if (hWaitTimer) {
	   KillTimer (hMain, hWaitTimer);
		hWaitTimer = 0;
	}	
	PostMessage (hMain, MYMSG_REFRESH, 0, 0);
   return sRC;
}   

//------------------------------------------------------------
// LoadProgram - Loads a program
//------------------------------------------------------------ 
INT LoadProgram (HWND hWnd, LPLINETOKEN pltProg, UINT wProgMax) {
   INT			i, sRC = 0, sErrCnt;
   HANDLE		hFileBuff;
   LPLINETOKEN	pltLine, pltOld;
	char			chChar;
   LPSTR			lpszSource;
   LPSTR			lpszLineStart;

	// Load file
	if (szCmdBuff[0] == '\0')
	   return 0;

	hFileBuff = GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, 65535);
	if (hFileBuff == 0) return ERR_OUT_OF_MEMORY;
	sRC = ReadFile (szCmdBuff, hFileBuff);
	if (sRC)	{
	   szCmdBuff[0] = 0;
		return sRC;
	}
	pltLine = pltProg;
	lpszSource = (LPSTR) GlobalLock (hFileBuff);
	if (lpszSource == 0)
	   return ERR_MEM_LOCK;

	SendDlgItemMessage (hWnd, IDD_LINELIST, LB_RESETCONTENT, 0, 0L);
	i = 0;
	sErrCnt = 0;
	while (*lpszSource != 0) {
	   pltOld = pltLine;
	   lpszLineStart = lpszSource;
	   sRC = TokenizeLine (&lpszSource, &pltLine, i++);
	   if (sRC) {
	      sErrCnt++;
	      if (sErrCnt > 100) {
	         sRC = 0;
	         break;
	      }   
	      PrintError (hWnd, i, sRC, FALSE);
	      sRC = 0;
		}
		chChar = *lpszSource;
		*lpszSource = 0;
		SendDlgItemMessage (hWnd, IDD_LINELIST, LB_ADDSTRING, 0, 
		                             (LPARAM) lpszLineStart);
		*lpszSource = chChar;
		if ((UINT)(((LPBYTE)pltLine-lpbTokLine) + MAXLINELEN) >= (UINT) TOKBUFFSIZE) {
	      PrintError (hWnd, i, ERR_FILE_TOO_COMPLEX, FALSE);
         sRC = ERR_FILE_TOO_COMPLEX;
		   break;
		}
	   if (*lpszSource != 0) lpszSource++;
	}
	pltLine->ucLen = 2;
	pltLine->ucType = TTYPE_EOF;
   GlobalUnlock (hFileBuff);
   GlobalFree (hFileBuff);
	
	if (strrchr (szCmdBuff, '\\') != 0)
	   strcpy (szProgName, strrchr (szCmdBuff, '\\')+1);
	else if (strlen (szCmdBuff) < sizeof (szProgName) - 1)
	   strcpy (szProgName, szCmdBuff);

	if (strrchr (szProgName, '.') != 0)
		*strrchr (szProgName, '.') = 0;

	SetWindowText (hMain, szProgName);
   return sRC;
}
//------------------------------------------------------------
// TermInterpeter - Terminates the interpeter 
//------------------------------------------------------------ 
INT TermInterpeter (void) {
   INT	i;

	for (i = 0; i < sLibCount; i++)
	   FreeLibrary (hLib[i]);

	LocalUnlock (hVarBuff);
	LocalFree (hVarBuff);
	GlobalUnlock (hTokBuff);
	GlobalFree (hTokBuff);
	if (hToolHelp) {
		if (hHoldTask)
			((THUREGPROC)*lpfnNotifyUnRegister)(NULL);
		FreeLibrary (hToolHelp);
	}	
	if (lpfnTHCallBack)
		FreeProcInstance (lpfnTHCallBack);
   return 0;
}   
//------------------------------------------------------------
// SetRunning - Starts and stops the interpeter
//------------------------------------------------------------ 
INT SetRunning (BOOL bGo) {

	if (bGo) {
	   if (!bAtEnd) {
			bRunning = TRUE;
			//Clear timer if stop due to Delay function
			if (hWaitTimer) {
			   KillTimer (hMain, hWaitTimer);
				hWaitTimer = 0;
			}
			if (hTimer2) {
			   KillTimer (hMain, hTimer2);
				hTimer2 = 0;
			}
			//If not running because of ExeAndWait, remove toolhelp hook
			if (lpfnNotifyUnRegister && hHoldTask) {
				((THUREGPROC)*lpfnNotifyUnRegister)(NULL);
				hHoldTask = 0;
			}
		}
	} else
		bRunning = FALSE;
	SendMessage (hMain, MYMSG_REFRESH, 0, 0);
	return 0;
}
//------------------------------------------------------------
// ParseCmdLine - Parses the command line
//------------------------------------------------------------ 
INT ParseCmdLine (LPSTR lpCmdLine) {
	INT	i, sRC = 0;
	
	// Parse Command line switches		
	while ((*lpCmdLine == '/') || (*lpCmdLine == '-')) {
	   lpCmdLine++;
		switch (*lpCmdLine) {
		   case 'l':
		   case 'L':
		   	bAutoStart = TRUE;
		   	break;
		   default:
		   	return ERR_UNKNOWN_CMDSWITCH;
		}
		do {
		   lpCmdLine++;   	
		} while ((*lpCmdLine != '\0') && (*lpCmdLine <= ' '));
	}
	if (*lpCmdLine == 0) return 0;
	// Copy filename
   i = 0;
	while ((*lpCmdLine != 0) && (i < sizeof (szCmdBuff))) {
		while ((*lpCmdLine != 0) && (*lpCmdLine > ' ') &&
			    (*lpCmdLine != '/')) 
			szCmdBuff[i++] = *lpCmdLine++;
		
		szCmdBuff[i++] = 0;
		while ((*lpCmdLine != 0) && (*lpCmdLine <= ' '))
			*lpCmdLine++;		
	}							       
	szCmdBuff[i++] = 0;
	strupr (szCmdBuff);  //Will only cap filename
	if (bAutoStart == FALSE)
		bAutoStart = TRUE;
	else
	   bAutoStart = FALSE;
	return 0;
}
//------------------------------------------------------------
// AddString - Adds a string to the string list
//------------------------------------------------------------
PBYTE AddString (LPSTR lpszString, char *wLinkBack, UINT wDLen, INT *sRC) {
   UINT	wLen, i;
   PITEM	piItem;
   LPSTR	lpszPtr;
   
	if (wDLen == 0)
		wLen = lstrlen (lpszString);
	else
		wLen = wDLen;
	                	
	//If adding a null string, Don't add to string list
	if (wLen == 0)
	   return (PBYTE) szNull;
	//See if we have the room
	if (pbVarStrings + wLen + 5 >= (PBYTE) ptVarList) {
		*sRC = ERR_NO_TOKEN_SPACE;
		return 0;
	}
	piItem = (PITEM) pbVarStrings;
	piItem->wLen = wLen + 5;
	piItem->wLinkBack = (WORD) wLinkBack;
	lpszPtr = piItem->szStr;
	for (i=0; i < wLen; i++)
		*lpszPtr++ = *lpszString++;
	*lpszPtr++ = 0;

	pbVarStrings = (PBYTE) OFFSETOF (lpszPtr);
	return (PBYTE) piItem->szStr;
}   
   
//------------------------------------------------------------
// AppendString - Adds a string to the end of a string
//------------------------------------------------------------
INT AppendString (PTOKEN ptDest, PTOKEN ptSrc) {
   INT	sLen, sRC = 0;
   PITEM	piItem;
   PBYTE pbNewString;
   char	*pszOldString;

	sLen = strlen (GetTokenString (ptSrc));
	if (sLen == 0) return 0;
	if (strlen (GetTokenString (ptDest)) == 0) {
		AddString ((LPSTR) GetTokenString (ptSrc), 
		                       (char *)&ptDest->dwData, ptSrc->wData, &sRC);
		return sRC;
	}
	// See if enough room in list for extra bytes
	if (pbVarStrings + sLen >= (PBYTE) ptVarList) 
		return ERR_NO_TOKEN_SPACE;

	//Copy original string to end of list
	pbNewString = AddString ((LPSTR) ptDest->dwData, 
	                         (char *)&ptDest->dwData, ptDest->wData, &sRC);
	if (sRC) return sRC;

	// Delete original string
	pszOldString = (char *)ptDest->dwData;
	ptDest->dwData = (DWORD)(LPBYTE) pbNewString;
	DeleteString (pszOldString, (WORD) &ptDest->dwData);
	piItem = (PITEM) (OFFSETOF (ptDest->dwData) - 4);
	
	strcpy (((char *)piItem)+piItem->wLen-1, GetTokenString (ptSrc));
	piItem->wLen += sLen;
	pbVarStrings += sLen;
	return 0;
}   
   
//------------------------------------------------------------
// DeleteString - Deletes a string to the string list
//------------------------------------------------------------
INT DeleteString (char *szString, WORD wLinkBack) {
	PITEM piItem, piNext;
	PBYTE	pbSrc, pbDest;
	DWORD	*pdwData;
	UINT	i, wLen;

	if (szString == 0) return 0;
	if (szString == szNull) return 0;
	piItem = (PITEM) (szString-4);
	if (piItem->wLinkBack == wLinkBack)
			piItem->wLinkBack = 0xffff;

	// Compress String list
	piNext = piItem;
	(PBYTE) piNext += piNext->wLen;
	while ((PBYTE) piNext < pbVarStrings) {
   	while (((PBYTE) piNext < pbVarStrings) && 
   	       (piNext->wLinkBack == 0xffff)) {
			(PBYTE) piNext += piNext->wLen;
	   }
		if ((PBYTE) piNext < pbVarStrings) {
		   pbSrc = (PBYTE) piNext;
		   pbDest = (PBYTE) piItem;

		   wLen = piNext->wLen;
		   for (i = 0; i < wLen; i++)
		      *pbDest++ = *pbSrc++;
		   pdwData = (DWORD *) piItem->wLinkBack;
		   *pdwData = (DWORD)(LPSTR) &piItem->szStr;

		   if (pbSrc == pbVarStrings)
		      pbVarStrings = pbDest;

		   piItem = (PITEM) pbDest;
		   piNext = (PITEM) pbSrc;
		}
	}
	if ((PBYTE) piNext >= pbVarStrings)
		pbVarStrings = (PBYTE) piItem;
   return 0;
}   

//------------------------------------------------------------
// AddTokenL - Adds a token to the token list
// This routine allows for a FAR pointer to the name.
//------------------------------------------------------------ 
PTOKEN AddTokenL (LPSTR lpszName, BYTE ucType, UINT wData, DWORD dwData, LPINT sRC) {
   char	szBuff[MAXSTRLEN+1];
	INT	sLocRC = 0;
	PTOKEN	ptRet;
	
	lstrcpy (szBuff, lpszName);
	ptRet = AddToken (szBuff, ucType, wData, dwData, &sLocRC);
	*sRC = sLocRC;
	return ptRet;
}   

//------------------------------------------------------------
// AddTokenInit - Special version of Add Token that saves
// space when adding statement and operator tokens.
//------------------------------------------------------------ 
INT AddTokenInit (char *szName, BYTE ucType, UINT wData, DWORD dwData) {

	ptVarList--;  //Make room for new token
	//See if out of space
	if ((PBYTE) ptVarList <= pbVarStrings) {
		ptVarList++;
		return ERR_NO_TOKEN_SPACE;
		return 0;
	}
  	wVarCount++; //Inc token count
		   
	ptVarList->pszName = szName;
   ptVarList->ucNameLen = (BYTE) strlen (szName);
   ptVarList->ucType = ucType;
	ptVarList->wData = wData;
	ptVarList->dwData = dwData;
	return 0;
}   
//------------------------------------------------------------
// AddToken - Adds a token to the token list
// If the name ptr = 0, a temporary token is added to the list
//------------------------------------------------------------ 
PTOKEN AddToken (char *szName, BYTE ucType, UINT wData, DWORD dwData, INT *sRC) {
	INT	i;
	PSTR	pszTemp;
	//Check to see if name too long
	if (szName != 0) {
		i = strlen (szName);
		if (i > MAXSTRLEN) {
			*sRC = ERR_NAME_TOO_LONG;
			return 0;
		}
	} else
	   i = 0;
	ptVarList--;  //Make room for new token
	//See if out of space
	if ((PBYTE) ptVarList <= pbVarStrings) {
		ptVarList++;
		*sRC = ERR_NO_TOKEN_SPACE;
		return 0;
	}
  	wVarCount++; //Inc token count
  	// Add Name to string list
  	if (szName != 0) {
		pszTemp = AddString ((LPSTR)szName, (char *)&ptVarList->pszName, 0, sRC);
		if (*sRC) {
			wVarCount--;
			ptVarList++;
			return 0;
		}
	} else 
	   pszTemp = szNull;
	   
	ptVarList->pszName = pszTemp;
   ptVarList->ucNameLen = (BYTE) i;
   ptVarList->ucType = ucType;
	ptVarList->wData = wData;
	ptVarList->dwData = dwData;
      
	if (ucType == TTYPE_STR) {
	   if (dwData != 0) {
			ptVarList->dwData = (DWORD)(LPSTR) AddString ((LPSTR) dwData, 
			                               (char *)&ptVarList->dwData, 0, sRC);
			if (*sRC) {
			   if (ptVarList->ucNameLen)
				   DeleteString (ptVarList->pszName, (WORD)&ptVarList->pszName);
				wVarCount--;
				ptVarList++;
				return 0;
			}	
		}
	}	
	return ptVarList;
}   
//------------------------------------------------------------
// FindToken - Finds a token in the token list
//------------------------------------------------------------ 
PTOKEN FindToken (LPSTR lpszTokenName, INT sNameLen) {
	UINT		i;
	char		chChar;
   PTOKEN	ptList;
   BOOL		bFound = FALSE;

	if (sNameLen > 255) return 0;
	ptList = ptVarList;

	if (sNameLen == 0) {
		for (i = 0; i < wVarCount; i++) {
   	   if ((WORD)ptList->pszName == (WORD)OFFSETOF (lpszTokenName)) {
      	   bFound = TRUE;
	         break;
	      }
	   	ptList++;    //Moved this in 1.3 to fix bug.
   	}
	} else {
	   chChar = *(lpszTokenName+sNameLen);
	   *(lpszTokenName+sNameLen) = 0;
		for (i = 0; i < wVarCount; i++) {
		   if ((INT) ptList->ucNameLen == sNameLen) {
	   	   if (lstrcmp (ptList->pszName, lpszTokenName) == 0) {
	      	   bFound = TRUE;
		         break;
		      }
	   	}
		   ptList++;
		}
	   *(lpszTokenName+sNameLen) = chChar;
	}	
	if (bFound)
	   return ptList;
	else
	   return 0;
}	

//------------------------------------------------------------
// FindTokenByType - Finds the most recient token of a given type
//------------------------------------------------------------ 
PTOKEN FindTokenByType (BYTE ucType) {
   PTOKEN	ptList;
   BOOL		bFound = FALSE;
   UINT      i;

	ptList = ptVarList;
	for (i = 0; i < wVarCount; i++) {
  	   if (ptList->ucType == ucType) {
      	   bFound = TRUE;
	         break;
   	}
	   ptList++;
	} 
	if (bFound)
	   return ptList;
	else
	   return 0;
}	

//------------------------------------------------------------
// TokFuncCallback - External entry point
//------------------------------------------------------------ 
DWORD CALLBACK TokFuncCallback (INT sFunction, LPTOKEN lptDest, 
               LPSTR lpszName, BYTE ucNewType, UINT wData, DWORD dwData) {
	INT	sRC = 0;
	PTOKEN	ptDest;
	LPIRQFUNC lpfnIRQTemp;
	
	ptDest = (PTOKEN)LOWORD(lptDest);
	switch (sFunction) {
	   case CBF_ADDTOKEN:
			AddTokenL (lpszName, ucNewType, wData, dwData, &sRC);
			return (DWORD) sRC;

	   case CBF_FINDTOKEN:
			return (DWORD) (LPTOKEN) FindToken (lpszName, wData);

	   case CBF_SETTOKEN:
			return (DWORD) SetToken (ptDest, ucNewType, wData, dwData);

	   case CBF_GETTOKENVAL:
			return GetTokenVal (ptDest);

	   case CBF_GETTOKENSTRING:
			return (DWORD)(LPSTR)GetTokenString (ptDest);

	   case CBF_RESOLVELINE:
			return (DWORD) ResolveLine ((LPLINETOKEN *)dwData, lpszName, wData);

		// The following callbacks were added in WinCmd 1.2
	   case CBF_GETCBMAXNUM:
			return (DWORD) CBF_LAUNCHLINE;

	   case CBF_GETWINCMDVER:
			return (DWORD) wWinCmdVer;

	   case CBF_EXECUTELINE:
			return (DWORD) ExecuteLine ((LPLINETOKEN *)dwData, (INT) wData);

	   case CBF_EXECUTE:
			return (DWORD) Execute ((LPLINETOKEN *)dwData, (INT) wData);
	
	   case CBF_SETIRQFUNC:
			lpfnIRQTemp = lpfnIRQ;
			lpfnIRQ = (LPIRQFUNC) dwData;			
			return (DWORD) (LPVOID)lpfnIRQTemp;
	
	   case CBF_CLEARHOLD:
			SetRunning (TRUE);
			return 0;

	   case CBF_PRINTERROR:
			return PrintError (hMain, (INT) wData, (INT)LOWORD (dwData),
			                           (BOOL) HIWORD (dwData));
	   case CBF_WRITEOUTPUT:
			return WriteOutput (hMain, lpszName);

		// The following callbacks were added in WinCmd 1.3
	   case CBF_GETHMAIN:
			return (LONG) hMain;

	   case CBF_DELTOKEN:
			return DeleteToken (ptDest);

	   case CBF_SETRUNNING:
			SetRunning (wData);
			return 0;

	   case CBF_LAUNCHLINE:
			return LaunchString (lpszName, wData);

		default:
			return ERR_BAD_FUNC;
	}		
}
//------------------------------------------------------------
// SetToken - Sets a token to a new value
//------------------------------------------------------------ 
INT SetToken (PTOKEN ptDest, BYTE ucNewType, UINT wData, DWORD dwData) {
   INT	sRC = 0;
   char	*pszOldString = 0;

	if ((ptDest->ucType != TTYPE_NUM) && (ptDest->ucType != TTYPE_STR))
	   return ERR_ILLEGAL_ASSIGNMENT;

	if ((ptDest->ucType != TTYPE_NUM) &&
	    ((char *)ptDest->dwData != ptDest->pszName))
		pszOldString = (char *) ptDest->dwData;

	ptDest->ucType = ucNewType;
	ptDest->wData = wData;
	ptDest->dwData = dwData;

	if (ucNewType == TTYPE_STR)
		ptDest->dwData = (DWORD)(LPSTR) AddString ((LPSTR) dwData, 
		                            (char *)&ptDest->dwData, 0, &sRC);

	if (sRC) 
	   ptDest->dwData = (DWORD)(LPSTR) ptDest->pszName;
	else 
	   if (pszOldString)
			DeleteString (pszOldString, (WORD) &ptDest->dwData);
	return sRC;
}	

//------------------------------------------------------------
// DeleteToken - Deltes a token in the token list
//------------------------------------------------------------ 
INT DeleteToken (PTOKEN ptToken) {

	if (ptToken->ucNameLen != 0) 
	   DeleteString (ptToken->pszName, (WORD)&ptToken->pszName);

	if ((ptToken->ucType == TTYPE_STR) &&
		 ((char *)ptToken->dwData != ptToken->pszName))
		DeleteString ((char *)ptToken->dwData, (WORD)&ptToken->dwData);
	
	if (ptToken == ptVarList) {
	   ptVarList++;
		wVarCount--;
	} else	
		ptToken->ucType = TTYPE_DEL;
   return 0;
}	
//------------------------------------------------------------
// GetTokenString - Returns a pointer to the string value
// for a token.
//------------------------------------------------------------ 
char *GetTokenString (PTOKEN ptDest) {
  
   switch (ptDest->ucType) {
      case TTYPE_NUM:
	     	ltoa (ptDest->dwData, szTemp, 10);
   	  	return szTemp;
  	  	
		case TTYPE_STR:
			if (ptDest->dwData != 0)
			   return (char *) ptDest->dwData;
			//If no string, fall through to default code
      default:
      	if (ptDest->ucNameLen == 0) {
      	   szTemp[0] = 0;
				return szTemp;
			} else	      	   
			   return ptDest->pszName;
	}
}  	
//------------------------------------------------------------
// GetTokenVal - Returns a pointer to the numberic value
// for a token.
//------------------------------------------------------------ 
LONG GetTokenVal (PTOKEN ptDest) {
  
	if (ptDest->ucType == TTYPE_NUM)
     	return ptDest->dwData;
	else
		return atol ((char *)GetTokenString (ptDest));
}  	

//------------------------------------------------------------
// FindChar - finds the next non-space character
//------------------------------------------------------------ 
LPSTR FindChar (LPSTR lpszLine) {

   while (*lpszLine != '\0') {
      if (*lpszLine == '\n') {
         break;
      }   
      if (*lpszLine == 0x0d) {
         *lpszLine = ' ';
      }   
      if (*lpszLine > ' ') break;
      lpszLine++;
   }
   return lpszLine;   
}   
//------------------------------------------------------------
// ComputeNum - Converts token to number
//------------------------------------------------------------ 
LONG ComputeNum (LPSTR *lpszLine, INT *sRC) {
   LONG lNum = 0, lTemp, lOld;
   INT	sBase = 10;
   BYTE	ucChar;

	if (**lpszLine == '0')
		switch (*(*lpszLine+1)) {
			case 'X':
	   		sBase = 16;
			   (*lpszLine) += 2;
	   		break;
			case 'O':
				sBase = 8;
			   (*lpszLine) += 2;
				break;
			case 'B':
				sBase = 2;
			   (*lpszLine) += 2;
				break;
			default:
				sBase = 10;
		}			   	
	ucChar = **lpszLine;	   	
  	while ((ucChar >= '0' && ucChar <= '9') ||
  	       ((ucChar >= 'A' && ucChar <= 'F') && sBase == 16)) {
		if (ucChar <= '9')
			lTemp = (LONG) ucChar - '0';
		else
			lTemp = (LONG) (ucChar - 'A' + 10);
		if (lTemp >= sBase)
			break;
		lOld = lNum;
   	lNum = (lNum * sBase) + lTemp;
	   (*lpszLine)++;
	   ucChar = **lpszLine;
	   if (sBase == 10) {
	   	if (lOld > lNum) {
   		   *sRC = ERR_OVERFLOW;
   		   break;
	   	}   
   	} else {
	   	if ((DWORD) lOld > (DWORD) lNum) {
   		   *sRC = ERR_OVERFLOW;
   		   break;
	   	}
	   }	
  	}
  	return lNum;
}  	
//------------------------------------------------------------
// UCaseLine - Converts line to uppercase.  Respects quoted
// strings.
//------------------------------------------------------------ 
void UCaseLine (LPSTR lpszLine) {
	BOOL	bInQuote = FALSE;

   while (*lpszLine != '\0' && *lpszLine != '\n') {
      if (!bInQuote) {
         if (*lpszLine >= 'a' && *lpszLine <= 'z')
	         *lpszLine = *lpszLine & (char) 0xdf;
		}
		if (*lpszLine == '"')
		   bInQuote = !bInQuote;
		lpszLine++;	
	}
	return;
}	
//------------------------------------------------------------
// TokenizeLine - Tokenizes the line.
//------------------------------------------------------------ 
INT TokenizeLine (LPSTR *plpszLineIn, LPLINETOKEN *ppbTokOut, UINT wLineNum) {
	LPSTR				lpszLine;
	LPSTR				lpszTokStart;
	LONG				lNum;
	char				chChar;
	BYTE				ucNameLen;
	INT				sRC = 0;
	BOOL				bLabel;
	LPLINETOKEN		pltOut;
	LPLINESTARTTOK	plstStart;
   PTOKEN			ptOldTok;

	lpszLine = *plpszLineIn;
	pltOut = *ppbTokOut;

	plstStart = (LPLINESTARTTOK) pltOut;
	plstStart->ucLen = 6;
	plstStart->ucType = TTYPE_LINE;
	plstStart->wLineNum = wLineNum;
	(PBYTE) pltOut += 6;

	UCaseLine (lpszLine);
	lpszLine = FindChar (lpszLine);

	while ((*lpszLine != '\n') && (*lpszLine != '\0') && (sRC == 0)) {
		lpszTokStart = lpszLine;
		ucNameLen = 0;
	   bLabel = FALSE;
		//Check for number
		if ((*lpszTokStart >= '0') && (*lpszTokStart <= '9')) {
			lNum = ComputeNum (&lpszLine, &sRC);
			if (sRC) break;
			if ((lNum > 32767) || (lNum < -32767)) {
				pltOut->ucType = TTYPE_LCONST;
				pltOut->ucLen = 6;
				((LPVARLCONST) pltOut)->lData = lNum;
				(PBYTE) pltOut += 6;
			} else {
				pltOut->ucType = TTYPE_SCONST;
				pltOut->ucLen = 4;
				((LPVARSCONST) pltOut)->sData = (INT) lNum;
				(PBYTE) pltOut += 4;
			}
		//check for quoted string
		} else if (*lpszTokStart == '"') {
		   lpszLine++;
		   lpszTokStart++;
			do {
				ucNameLen++;
				lpszLine++;
			} while ((*lpszLine != '\n') && (*lpszLine != '\0') &&
			         (*lpszLine != '"') && (ucNameLen < MAXSTRLEN));
			if (*lpszLine == '"') {
				pltOut->ucType = TTYPE_STRCONST;
				pltOut->ucLen = (BYTE) (ucNameLen + 3);
				chChar = *lpszLine;
				*lpszLine = '\0';
				lstrcpy ((LPSTR) &pltOut->ptToken, lpszTokStart);
				*lpszLine = chChar;
				(PBYTE) pltOut += ucNameLen + 3;
			   lpszLine++;
			} else
				sRC = ERR_NO_END_QUOTE;
		//Otherwise it can be tokenized	
		} else {
		   //See if allowable variable
			if (((*lpszTokStart >= 'A') && (*lpszTokStart <= 'Z')) ||
		         (*lpszTokStart == '_')                            ) {
				do {
					ucNameLen++;
					lpszLine++;
				} while ((*lpszLine != '\n') &&  (*lpszLine != '\0') &&
	  			         (((*lpszLine >= 'A') && (*lpszLine <= 'Z')) ||
					       ((*lpszLine >= '0') && (*lpszLine <= '9')) ||
				          (strchr ("_.$\\:", *lpszLine) != 0)      ));
				if (*(lpszLine-1) == ':') {
				   bLabel = TRUE;
				   ucNameLen--;
					lpszLine--;
					*lpszLine = ' ';
				}   				          
			} else {
			   //Must be operator, or other string constant
				do {
					ucNameLen++;
					lpszLine++;
				} while ((ucNameLen < 2) && (*lpszLine != '\0')    &&
				         (*lpszLine != '_')                        &&
				         (((*lpszLine > ' ') && (*lpszLine < '0')) ||
					       ((*lpszLine > '9') && (*lpszLine < 'A')) ||
				          (*lpszLine > 'Z')                      ));
				          
				ptOldTok = FindToken (lpszTokStart, 2);
				if ((ptOldTok == 0) && (ucNameLen == 2)) {
				   ucNameLen = 1;
				   lpszLine--;
				}   
			}
			ptOldTok = FindToken (lpszTokStart, ucNameLen);
			if (ptOldTok == 0) {
			   chChar = *lpszLine;
			   *lpszLine = 0;
				if (bLabel) 
				   ptOldTok = AddTokenL (lpszTokStart, TTYPE_LABEL, 0, 
				                         (DWORD)(LPVOID)pltOut, &sRC);
				else
				   ptOldTok = AddTokenL (lpszTokStart, TTYPE_STR, 0, 0, &sRC);
				*lpszLine = chChar;
				pltOut->ucType = TTYPE_VAR;
			} else {
			   // If comment, skip to end of line
				if (ptOldTok->ucType == TTYPE_CMT2EOL) {
					while ((*lpszLine != '\n') && (*lpszLine != '\0')) {
					   if (*lpszLine == 0x0d)
					      *lpszLine = ' ';
						lpszLine++;
					}	
				   break;
				}   
				switch (ptOldTok->ucType) {
					case TTYPE_NUM:
					case TTYPE_STR:
						if (bLabel) {
							ptOldTok->ucType = TTYPE_LABEL;
							ptOldTok->dwData = (DWORD)(LPVOID)pltOut;
						} else
							pltOut->ucType = TTYPE_VAR;
					   break;
					case TTYPE_LABEL:
						if (bLabel) 
						   sRC = ERR_MULTI_DEF_LABEL;
						pltOut->ucType = TTYPE_VAR;
						break;
					default:
						pltOut->ucType = ptOldTok->ucType;
				}		
			} 
			if (!bLabel) {
			   pltOut->ucLen = 4;
				pltOut->ptToken = ptOldTok;
			   (PBYTE) pltOut += 4;
			}   
		}			
		lpszLine = FindChar (lpszLine);
	}

   if ((*lpszLine != '\n') || (*lpszLine != '\0')) 
	   while ((*lpszLine != '\n') && (*lpszLine != '\0')) 
	      lpszLine++;

	pltOut->ucLen = 2;
	pltOut->ucType = TTYPE_EOL;
   (PBYTE) pltOut += 2;

	plstStart->wLineLen = (LPBYTE) pltOut - (LPBYTE) plstStart;

	*plpszLineIn = lpszLine;
	*ppbTokOut = pltOut;

	return sRC;
}	
//------------------------------------------------------------
//------------------------------------------------------------
INT lstrcspn (LPSTR lpIn, LPSTR lpTemplate) {
	INT i, sLen;
	BOOL fBreak = FALSE;
	LPSTR lpStart;
	
	sLen = lstrlen (lpTemplate);
	lpStart = lpIn;
	while ((*lpIn != '\0') && !fBreak) {
		for (i = 0; i < sLen && !fBreak; i++)
			if (*lpIn == *(lpTemplate+i))
				fBreak = TRUE;
		if (!fBreak)
			lpIn++;		
	}
	return (INT)(lpIn - lpStart);
}	
//------------------------------------------------------------
//------------------------------------------------------------
LPSTR lstrrchr (LPSTR lpIn, char ch) {
	LPSTR lpStart;

	if (*lpIn == '\0')
		return 0;	
	lpIn = lpIn + lstrlen (lpIn) - 1;

	while (lpIn >= lpStart) 
		if (*lpIn != ch) 
			return lpIn;
		else	
			lpIn--;
	return 0;
}	
//------------------------------------------------------------
// LaunchString - Launch a program
//------------------------------------------------------------ 
LONG LaunchString (LPSTR lpszProg, UINT usShowWin) {
   char	szTemp[MAXLINELEN + 12], chChar;
   LPSTR lpszExt;
   char	*pszDefExt;
   INT	sRC, i, j;
   PTOKEN	ptName;
   OFSTRUCT	of;
   BOOL	bFound;

	if (*(lpszProg+1) == ':')
		i = lstrcspn (lpszProg+2, " *+,/:;<>=|") + 2;
	else
		i = lstrcspn (lpszProg, " *+,/:;<>=|");
	chChar = *(lpszProg+i);
	*(lpszProg+i) = '\0';
	AnsiUpper (lpszProg);
   bFound = FALSE;
   ptName = FindToken (lpszProg, i);
   if ((ptName != 0) && (ptName->ucType == TTYPE_INTCMD)) {
	   strcpy (szTemp, getenv("COMSPEC"));
	   strcat (szTemp, " /C ");
		*(lpszProg+i) = chChar;
		lstrcpyn (&szTemp[strlen(szTemp)], lpszProg, 
		          sizeof (szTemp) - strlen (szTemp));
		lpszProg = szTemp;
		if (usShowWin == 0)
			usShowWin = SW_SHOWMINNOACTIVE;
		bFound = TRUE;
	} else {	
		lpszExt = lstrrchr (lpszProg, '.');
		if ((lpszExt == 0) || (lpszExt < lstrrchr(lpszProg, '\\'))) {
		   //Make room on cmdline for ext
			for (j = lstrlen (lpszProg+i+1)+3; j > 0; j--) 
				*(lpszProg+i+j+4) = *(lpszProg+i+j);
   	
			*(lpszProg+i) = '.';
			*(lpszProg+4) = 0;
			pszDefExt = szProgExt;
			bFound = FALSE;
			while (*pszDefExt != 0) {
			   for (j = 0; j < 3; j++)
					*(lpszProg+i+j+1) = *pszDefExt++;
			
				if (OpenFile (lpszProg, &of, OF_EXIST) != HFILE_ERROR) {
					bFound = TRUE;
					break;
				}	
				while ((*pszDefExt <= ' ') && (*pszDefExt != 0))
					pszDefExt++;
			}
			*(lpszProg+i+4) = chChar;
		} else {
			*(lpszProg+i) = chChar;
			bFound = TRUE;
		}	
		if (usShowWin == 0)
			usShowWin = SW_SHOWNORMAL;
	}	
	if (bFound) {
		sRC = WinExec (lpszProg, usShowWin);
		if (sRC < 32)
			return MAKELONG (0, ERR_LAUNCH_ERR - sRC);
		else
			return MAKELONG (sRC, 0);
	} else
		return MAKELONG (0, ERR_FOPEN_ERR - of.nErrCode);
}
//------------------------------------------------------------
// ResolveLine - Converts line tokens to text
//------------------------------------------------------------ 
INT ResolveLine (LPLINETOKEN *pltCurrToken, LPSTR lpszOutput, INT sMaxSize) {
   PTOKEN ptTemp;
   INT	sRC = 0;

	*lpszOutput = '\0';
	while ((*pltCurrToken)->ucType != TTYPE_EOL) {
		ptTemp = AddToken (0, TTYPE_NUM, 0, 0, &sRC);
		if (sRC) break;
		sRC = Eval (pltCurrToken, ptTemp, FUNC_EXECUTE);
		if (sRC) {
			DeleteToken (ptTemp);
			break;
		}
		lstrcat (lpszOutput, (LPSTR)GetTokenString (ptTemp));
		DeleteToken (ptTemp);
		lstrcat (lpszOutput, " ");
	}
	*(lpszOutput+lstrlen(lpszOutput)-1) = '\0';
	return sRC;
}
//------------------------------------------------------------
// SkipLine - Scan until start of next line
//------------------------------------------------------------ 
LONG SkipLine (LPLINETOKEN *pltStart) {

   while ((*pltStart)->ucType != TTYPE_EOL)
		*(PWORD)pltStart += (*pltStart)->ucLen;
	*(PWORD)pltStart += (*pltStart)->ucLen;
	return 0;
}
//------------------------------------------------------------
// WriteOutput - Writes a line to the output listbox
//------------------------------------------------------------ 
void WriteOutput (HWND hWnd, LPSTR pszOutput) {
   LONG	index;
   
	SendDlgItemMessage (hWnd, IDD_OUTLIST, WM_SETREDRAW, FALSE, 0);
	if (SendDlgItemMessage (hWnd, IDD_OUTLIST, LB_GETCOUNT, 0, 0) > MAXOUTLINES)
		SendDlgItemMessage (hWnd, IDD_OUTLIST, LB_DELETESTRING, 0, 0);
   
	index = SendDlgItemMessage (hWnd, IDD_OUTLIST, LB_ADDSTRING, 0, 
	                            (LPARAM)(LPSTR) pszOutput);
	SendDlgItemMessage (hWnd, IDD_OUTLIST, LB_SETTOPINDEX, (WORD)index, 0);
	SendDlgItemMessage (hWnd, IDD_OUTLIST, WM_SETREDRAW, TRUE, 0);
	                    
	if (IsIconic (hWnd)) {
		bShowOut = TRUE;		                    
		ShowWindow (hWnd, SW_RESTORE);
	} else if (!bShowOut) {
	   bShowOut = TRUE;
	   PostMessage (hWnd, MYMSG_REFRESH, 0, 0);
	}
	bAutoStart = FALSE;
	return;	                    
}	                    

//------------------------------------------------------------
// GetErrorString - Returns an error string 
//------------------------------------------------------------ 
BOOL GetErrorString (INT sLineNum, INT sErrorCode, char *pszOut, INT sMax) {
	INT	rc;

	strcpy (pszOut, "Error - ");
	if (sLineNum != -1) {
		strcat (pszOut, "Line: ");
		itoa (sLineNum+1, szTemp, 10);
		strcat (pszOut, szTemp);
		strcat (pszOut, "     ");
	}	
	if (sErrorCode > 0)
	   sErrorCode += 11;
	else 
	   sErrorCode = abs (sErrorCode);
	// WinCmd 1.2 - Added check for Lib error codes		
	if (sErrorCode > -ERR_LIBERROR_START) {
		sErrorCode += ERR_LIBERROR_START;
		rc = ((LPLIBGETERR)*(lpfnLibErr[sErrorCode/MAXLIBERRS]))
 				           (sErrorCode%MAXLIBERRS, 
 				           pszOut + strlen(pszOut), sMax - strlen (pszOut));
		if (rc == 0) {
			itoa (sErrorCode, szTemp, 10);
			strcat (pszOut, "Library Error: ");
			strcat (pszOut, szTemp);
		}			
	} else if (LoadString (hInst, sErrorCode, pszOut + strlen(pszOut), 
	                       sMax - strlen (pszOut)) == 0) {
		itoa (sErrorCode, szTemp, 10);
		strcat (pszOut, "Error number: ");
		strcat (pszOut, szTemp);
	}
	return TRUE;
}	                    

//------------------------------------------------------------
// PrintError - Displays an error message on the output listbox
//------------------------------------------------------------ 
void PrintError (HWND hWnd, INT sLineNum, INT sErrorCode, BOOL bSetList) {
   char	szTempStr[MAXLINELEN];

	GetErrorString (sLineNum, sErrorCode, szTempStr, sizeof (szTempStr));
	WriteOutput (hWnd, szTempStr);
	if (bSetList) {
		bShowList = TRUE;
		bShowOut = TRUE;
	   PostMessage (hWnd, MYMSG_REFRESH, 0, 0);
	   SetActiveWindow (hWnd);
	}   
	return;	                    
}	                    

//----------------------------------------------------------------------
// Read File - Allocates a buff in global memory and reads in a file
//----------------------------------------------------------------------
INT ReadFile (char *szFileName, HANDLE hFileBuff) {
   
	OFSTRUCT	ofFile;
	HFILE		hFileHandle;
	UINT		wBytesRead;
	LPBYTE 	lpbFileBuff;

	lpbFileBuff = GlobalLock (hFileBuff);

	hFileHandle = OpenFile (szFileName, &ofFile, OF_READ);
	if (hFileHandle == -1) {
	   GlobalUnlock (hFileBuff);
	   return ERR_FOPEN_ERR - ofFile.nErrCode;
	}
	wBytesRead = _lread (hFileHandle, lpbFileBuff, (WORD)65530);
	_lclose (hFileHandle);

	if (wBytesRead == 65530) {
	   GlobalUnlock (hFileBuff);
	   return ERR_FILE_TOO_BIG;
	}
	*(lpbFileBuff + wBytesRead++) = 0x0d;
	*(lpbFileBuff + wBytesRead++) = 0x0a;
	*(lpbFileBuff + wBytesRead++) = 0;
	
   GlobalUnlock (hFileBuff);
   GlobalReAlloc (hFileBuff, wBytesRead+2, 0);
   return 0;
}
//------------------------------------------------------------
//	StuffVKey - Send a character to the keyboard driver
//------------------------------------------------------------ 
void WINAPI Keybd_Event (void);
void StuffVKey (BYTE ucVKey, BOOL bMake) {
	WORD	RegAX, RegBX;
	
	if (bMake)
		RegAX = (WORD) ucVKey;
	else	
		RegAX = 0x8000 | (WORD) ucVKey;
	RegBX = (WORD) MapVirtualKey (ucVKey, 0);
	_asm {
	   mov ax,RegAX
	   mov bx,RegBX
	}   
   Keybd_Event();
}
//------------------------------------------------------------
// StuffOneKey - Processes a character in a stuff key string
//------------------------------------------------------------ 
INT StuffOneKey (char **pszString) {
   BYTE	ucVKey, ucShift = 0;
   INT	sRC = 0, sRepCnt = 0, i;
   UINT  wKey;
   BOOL	bFound;
   char	*pszEnd;
   char	*pszNum;

   bFound = FALSE;
	ucVKey = **pszString;
	if (ucVKey == 0) return 0;
	(*pszString)++;
	switch (ucVKey) {
	   case '+':
			ucShift = VK_SHIFT;
			break;
	   case '%':
			ucShift = VK_MENU;
			break;
	   case '^':
			ucShift = VK_CONTROL;
			break;
	   case '(':
   		sRC = StuffKeyString (pszString);
   		if (sRC == 1)
   		   return 0;
			else if (sRC == 0)
			   return ERR_BAD_SENDKEY_STR;
			else
			   return sRC;
	   case ')':
   		return 1;

	   case '{':
	   	if (*(*pszString+1) == '}') {
				ucVKey = **pszString;
				(*pszString) += 2;
			} else {
			   pszEnd = strchr (*pszString, '}');
			   if (pszEnd == 0) return ERR_BAD_SENDKEY_STR;
			   *pszEnd = 0;
			   pszNum = strchr (*pszString, ' ');
			   if (pszNum != 0) *pszNum = 0;
				for(i = 0; i < dim(SKList); i++) {
					if (strcmpi (SKList[i].szName, *pszString) == 0) {
					   bFound = TRUE;
					   break;
					}   
				}
				if (!bFound) return ERR_BAD_SENDKEY_STR;
				ucVKey = SKList[i].ucKey;
				if (pszNum != 0) {
					*pszNum++ = ' ';
					(*pszString) = pszNum;
					sRepCnt = atoi(pszNum);
				}
				*pszEnd++ = '}';
				(*pszString) = pszEnd;
			}   
   		break;

	}
	if (ucShift) { 
  		StuffVKey (ucShift, TRUE);
  		StuffOneKey (pszString);
  		StuffVKey (ucShift, FALSE);
	} else {
	   if (!bFound) {
	  		wKey = VkKeyScan (ucVKey);
	  		ucVKey = (BYTE) wKey;
	  	} else wKey = 0;	
  		if (HIBYTE (wKey) == 1) {
	  		StuffVKey (VK_SHIFT, TRUE);
	  		StuffVKey (ucVKey, TRUE);
	  		StuffVKey (ucVKey, FALSE);
	  		StuffVKey (VK_SHIFT, FALSE);
		} else {
	  		StuffVKey (ucVKey, TRUE);
			StuffVKey (ucVKey, FALSE);
		}	   
	}   
	return sRC;
}		
//------------------------------------------------------------
// StuffKeyString - Stuffs a 'stuff key' string into the
// system event queue.
//------------------------------------------------------------ 
INT StuffKeyString (char **pszString) {
   INT	sRC = 0;
	
	while ((**pszString != 0) && (sRC == 0)) 
		sRC = StuffOneKey (pszString);
	return sRC;
}		
//------------------------------------------------------------
// DropFiles - Processes drop of files on a window
//------------------------------------------------------------ 
INT DropFiles (HWND hWnd, HANDLE hDrop) {
	LPSTR lpszData;
	INT i, sFiles;
	UINT usSize = 0, usTotal;
	
	if (hDropData)
		GlobalFree (hDropData);
		
	sFiles = DragQueryFile (hDrop, -1, 0, 0);
	
	usTotal = min ((sFiles * 128) + 4, 0xffff);
	hDropData = GlobalAlloc (GHND, usTotal);
	if (hDropData == 0)
		return 0;
	lpszData = GlobalLock (hDropData);
	*((LPINT)lpszData)++ = sFiles;
	//Read all file names into a buffer.
	for (i = 0; i < sFiles; i++) {
		usSize += DragQueryFile (hDrop, i, lpszData, usTotal - usSize);
		lpszData += lstrlen (lpszData) + 1;
	}
	DragFinish (hDrop);
	GlobalUnlock (hDropData);
	
	fIntFlags |= IFLAG_DROPFILES;
	SetRunning (TRUE);
	return 0;
}		
//------------------------------------------------------------
// CallFunc - Calls user defined and built in functions
//------------------------------------------------------------ 
INT CallFunc (LPLINETOKEN *pltLine, PTOKEN ptCurrValue, INT sFunc) {
   LPLINETOKEN	pltFunc;
   PTOKEN	ptInd;
   PTOKEN	ptNextValue, ptArg1;
   INT		sRC = 0, sTokCnt;
   UINT		wTemp;

	ptInd = (*pltLine)->ptToken;
	sTokCnt = (*pltLine)->ucLen;
	*(PWORD)pltLine += sTokCnt;
	// If no open paren, just return func name as string
	if ((*pltLine)->ucType != TTYPE_OPAREN) 
		return SetToken (ptCurrValue, TTYPE_STR, 0, 
		                 (DWORD)(LPSTR)ptInd->pszName);
	sTokCnt = (*pltLine)->ucLen;
	*(PWORD)pltLine += sTokCnt;

	// Push arguments onto var list
	ptArg1 = AddToken (0, TTYPE_STR, 0, (DWORD)(LPSTR)ptInd->pszName, &sRC);
	if (sRC) return sRC;
	sTokCnt = 1;
	while (((*pltLine)->ucType != TTYPE_CPAREN) &&
	       ((*pltLine)->ucType != TTYPE_EOL) && (sRC == 0)) {
		ptNextValue = AddToken (0, TTYPE_NUM, 0, 0, &sRC);
  		sRC = Eval (pltLine, ptNextValue, sFunc);
  		sTokCnt++;
	}
	if ((sRC == 0) && (sFunc == FUNC_EXECUTE)) {
		switch (ptInd->ucType) {
			case TTYPE_LABEL:
				SetToken (ptCurrValue, TTYPE_RETVAL, sTokCnt, (DWORD)(LPTOKEN)ptArg1);
				ptCurrValue->ucType = TTYPE_RETVAL;
		   	pltFunc = (LPLINETOKEN) ptInd->dwData;
		   	wTemp = wCurrentLine;
				while (sRC == 0)
				   sRC = ExecuteLine (&pltFunc, FUNC_EXECUTE);
				if (sRC == RC_END_OF_SUB) {
					sRC = 0;
					wCurrentLine = wTemp;
				}   
				break;
			case TTYPE_LOCALFUNC:
				SetToken (ptCurrValue, TTYPE_NUM, sTokCnt, (DWORD)(LPTOKEN)ptArg1);
				sRC = (*((PFUNCTOK)ptInd)->Fxn)(pltLine, ptCurrValue);
				break;
			case TTYPE_LIBFUNC:
				SetToken (ptCurrValue, TTYPE_NUM, sTokCnt, (DWORD)(LPTOKEN)ptArg1);
				sRC = ((LPLIBFUNC)*(lpfnLibFunc[HIWORD(ptInd->dwData)]))
 				           (pltLine, (LPTOKEN)ptCurrValue, LOWORD (ptInd->dwData),
				            lpfnTokenFuncCallback, hMain);
				break;
			default:
				sRC = ERR_FUNC_NOT_FOUND;
				break;

		}
	}	
	// Remove arg tokens from var list
	while (--sTokCnt >= 0)
	   DeleteToken (ptArg1-sTokCnt);
	if (sRC) return sRC;
	if ((*pltLine)->ucType != TTYPE_CPAREN) 
	   return ERR_NO_CLOSE_PAREN;
	sTokCnt = (*pltLine)->ucLen;
	*(PWORD)pltLine += sTokCnt;
	return sRC;
}

//------------------------------------------------------------
// GetOp - Recursive routine that evaluates an expression
//------------------------------------------------------------ 
INT GetOp (LPLINETOKEN *pltLine, PTOKEN ptCurrValue, UINT *wLevel, INT sFunc) {
   UINT		oldLevel, wNewLev;
   INT		sRC = 0;
   BYTE		ucTokLen;
   POPTOK	potTokIndirect;
   PTOKEN	ptInd;
   PTOKEN	ptNextValue;

	ucTokLen = (*pltLine)->ucLen;
	switch ((*pltLine)->ucType) {
		case TTYPE_SCONST:
			sRC = SetToken (ptCurrValue, TTYPE_NUM, 0,
			                (DWORD)((LPVARSCONST) *pltLine)->sData);
			*(PWORD)pltLine += ucTokLen;
			break;

		case TTYPE_LCONST:
			sRC = SetToken (ptCurrValue, TTYPE_NUM, 0,
			                (DWORD)((LPVARLCONST) *pltLine)->lData);
			*(PWORD)pltLine += ucTokLen;
			break;

		case TTYPE_STRCONST:
			sRC = SetToken (ptCurrValue, TTYPE_STR, 0,
			                (DWORD)(LPSTR)((LPSTRCONST)*pltLine)->szData);
			*(PWORD)pltLine += ucTokLen;
			break;

		case TTYPE_VAR:
			ptInd = (PTOKEN) (*pltLine)->ptToken;
			switch (ptInd->ucType) {
				case TTYPE_LABEL:
			   sRC = CallFunc (pltLine, ptCurrValue, sFunc);
			   break;
				case TTYPE_NUM:
               sRC = SetToken (ptCurrValue, ptInd->ucType, ptInd->wData, 
				                   GetTokenVal(ptInd));
				  	*(PWORD)pltLine += ucTokLen;
				   break;
				case TTYPE_STR:
               sRC = SetToken (ptCurrValue, ptInd->ucType, ptInd->wData, 
				                   (DWORD)(LPSTR)GetTokenString(ptInd));
				  	*(PWORD)pltLine += ucTokLen;
				   break;
				default:				   
			   	sRC = ERR_SYNTAX;
			   	break;
			}	
			break;
			
		case TTYPE_OPAREN:  			
  			ptInd = (*pltLine)->ptToken;
  			*(PWORD)pltLine += ucTokLen;
			if (sFunc == FUNC_EXECUTE) {
				sRC = (*((PFUNCTOK)ptInd)->Fxn)(pltLine, ptCurrValue);
			} else {
				Eval2Num (pltLine, sFunc, &sRC);
				if (sRC) return sRC;
				if ((*pltLine)->ucType != TTYPE_CPAREN) 
			      return ERR_NO_CLOSE_PAREN;
			   else
					*(PWORD)pltLine += (*pltLine)->ucLen;
         }
			break;

		case TTYPE_LIBFUNC:
		case TTYPE_LOCALFUNC:
		   sRC = CallFunc (pltLine, ptCurrValue, sFunc);
			break;

		case TTYPE_LUOP:
  			potTokIndirect = (POPTOK) (*pltLine)->ptToken;
			*(PWORD)pltLine += ucTokLen;

			//Kluge to allow minus sign to have 2 levels, 13 for
			//reguar op and 15 for unary op.			
	  		if (potTokIndirect->wLevel == 13)
		  		wNewLev = 15;
			else		  		
		  		wNewLev = potTokIndirect->wLevel;
	  		if (*wLevel >= wNewLev) {
				*wLevel = wNewLev;
				return 0;
			} else {
  			   oldLevel = *wLevel;
				*wLevel = wNewLev;

				ptNextValue = AddToken (0, TTYPE_NUM, 0, 0, &sRC);
				if (sRC) return sRC;				
	   		sRC = GetOp (pltLine, ptNextValue, wLevel, sFunc);
				if (sRC) return sRC;
				sRC = (*potTokIndirect->Fxn)(pltLine, ptCurrValue, ptNextValue);
				DeleteToken (ptNextValue);
				*wLevel = oldLevel;
			}
			break;

		default:
			if ((*pltLine)->ucType < TTYPE_EOL) {
				ptInd = (PTOKEN) (*pltLine)->ptToken;
				if (ptInd->ucNameLen != 0) {
				   SetToken (ptCurrValue, TTYPE_STR, 0, (DWORD)(LPSTR)GetTokenString(ptInd));
					*(PWORD)pltLine += ucTokLen;
					break;
				}
		   	sRC = ERR_EXPECTED_NUMBER;
			}
		   sRC = ERR_EXPECTED_NUMBER;
	}	
	if (sRC) return sRC;
	oldLevel = (*wLevel)++;
	while (oldLevel < *wLevel) {
	   *wLevel = oldLevel;
		ucTokLen = (*pltLine)->ucLen;
	   
		if ((*pltLine)->ucType == TTYPE_EOL) {
		   *wLevel = 0;
	   	return 0;
		} else {
  			if (((*pltLine)->ucType == TTYPE_LOP) || 
  			    ((*pltLine)->ucType == TTYPE_LUOP) ||
  			    ((*pltLine)->ucType == TTYPE_ASSIGN_CMP)) {

	  			potTokIndirect = (POPTOK)(*pltLine)->ptToken;
  				if (*wLevel >= potTokIndirect->wLevel) {
					*wLevel = potTokIndirect->wLevel;
					return 0;
				} else {
					*wLevel = potTokIndirect->wLevel;
					*(PWORD)pltLine += ucTokLen;

					ptNextValue = AddToken (0, TTYPE_NUM, 0, 0, &sRC);
					if (sRC) return sRC;
	   			sRC = GetOp (pltLine, ptNextValue, wLevel, sFunc);
					if (sRC) return sRC;
					sRC = (*potTokIndirect->Fxn)(pltLine, ptCurrValue, ptNextValue);
					DeleteToken (ptNextValue);
				}
			} else if ((*pltLine)->ucType == TTYPE_OPAREN) {
				sRC = ERR_BAD_FUNC_NAME;
				break;
			} else if ((*pltLine)->ucType == TTYPE_BREAK) {
				*(PWORD)pltLine += (*pltLine)->ucLen;
			   *wLevel = 0;
			} else {
			   *wLevel = 0;
  			}   
		}		
	}           	
	return sRC;
}
//------------------------------------------------------------
// Eval - Returns a token for an expression
//------------------------------------------------------------ 
INT Eval (LPLINETOKEN *pltLine, PTOKEN ptOut, INT sFunc) {
   UINT 	wLevel = 1;
	return  GetOp (pltLine, ptOut, &wLevel, sFunc);
}	

//------------------------------------------------------------
// Eval2Num - Returns a number value for an expression
//------------------------------------------------------------ 
LONG Eval2Num (LPLINETOKEN *pltLine, INT sFunc, INT *sRC) {
   UINT 		wLevel = 1;
   LONG		lResult;
   PTOKEN	ptTemp;
   
	ptTemp = AddToken (0, TTYPE_NUM, 0, 0, sRC);
	if (*sRC) return 0;
	*sRC = GetOp (pltLine, ptTemp, &wLevel, sFunc);
	lResult = GetTokenVal (ptTemp);
	DeleteToken (ptTemp);
	return lResult;
}	

//------------------------------------------------------------
// GetStatementType - Returns type of next token
//------------------------------------------------------------ 
BYTE GetStatementType (LPLINETOKEN pltNext) {

	if (pltNext->ucType == TTYPE_LINE)
	  	(PBYTE)pltNext += pltNext->ucLen;
	if (pltNext->ucType == TTYPE_CMT2EOL) {
	   while (pltNext->ucType != TTYPE_EOL)
			(PBYTE)pltNext += pltNext->ucLen;
		(PBYTE)pltNext += pltNext->ucLen;
		return GetStatementType (pltNext);
	}	
	return pltNext->ucType;
}   
//------------------------------------------------------------
// ExecuteLine - Evaluates and calls the proper routine to
// execute a line in the file.
//------------------------------------------------------------ 
INT ExecuteLine (LPLINETOKEN *pltCurrent, INT sFunc) {
   INT	sRC = 0;
	LPIRQFUNC lpfnIRQTemp;
	LPLINETOKEN ltDrop;
	PTOKEN pToken;
	WORD wTemp;

	if (sFunc & FUNC_EXECUTE) {
		if (!bRunning) {
			if (!(sFunc & FUNC_NOREFRESH) && 
			    ((*pltCurrent)->ucType == TTYPE_LINE)) 
			   wCurrentLine = ((LPLINESTARTTOK) *pltCurrent)->wLineNum;

			PostMessage (hMain, MYMSG_REFRESH, 0, 0);
		}	
		//Our message loop is here! If we recieved a WM_QUIT, report it
		sFunc &= ~FUNC_NOREFRESH;
		if (!MyYield()) 
			return ERR_UNEXP_PROGTERM;

		if (bReset)
			return ERR_RESET;

		// If single stepping
		if (bSingleStep)
			SetRunning (FALSE);
		//
		// Hook added in WinCmd 1.2.  Modified in 1.3
		//
		if ((lpfnIRQ) && (sFunc == FUNC_EXECUTE))  {
			lpfnIRQTemp = lpfnIRQ;
			lpfnIRQ = 0;
			sRC = (*lpfnIRQTemp)(pltCurrent, sRC, lpfnTokenFuncCallback, hMain);
		}
		// Dropfile check added in 1.3
		if (fIntFlags & IFLAG_DROPFILES) {
			fIntFlags &= ~IFLAG_DROPFILES;
			pToken = FindToken ("_DROPPTR", 8);
			if (pToken) {
				if (pToken->dwData) {	
			   	ltDrop = (LPLINETOKEN) pToken->dwData;
			   	wTemp = wCurrentLine;
				   sRC = Execute (&ltDrop, FUNC_EXECUTE);
					if (sRC == 0)
						wCurrentLine = wTemp;
					else if (sRC > 1)
					   sRC = -(sRC + 11);
				}		
			}	
		}
	}
	if (sRC == 0)
		sRC = Execute (pltCurrent, sFunc);  //Execute a line
   return sRC;
}   
//------------------------------------------------------------
// Execute - Processes a series of tokens.
//------------------------------------------------------------ 
INT Execute (LPLINETOKEN *pltStart, INT sFunc) {
	LPLINETOKEN		pltCurr, pltNext;
	PSTATEMENTTOK	pstTokIndirect;
	INT				sRC = 0;
	PTOKEN			ptTemp;

	pltCurr = *pltStart;

	if (pltCurr->ucType == TTYPE_LINE) {
	   wCurrentLine = ((LPLINESTARTTOK) pltCurr)->wLineNum;
  		(PBYTE) pltCurr += pltCurr->ucLen;
	}
  	(LPBYTE) pltNext = (LPBYTE) pltCurr + pltCurr->ucLen;
	switch (pltCurr->ucType) {

   	case TTYPE_INTCMD:
   	case TTYPE_STRCONST:
		   sRC = LocalLaunch (&pltCurr, sFunc);
		   *pltStart = pltCurr; 
	   	break;

	   case TTYPE_VAR:
			if (sFunc == FUNC_EXECUTE) {
		   	if ((pltNext->ucType == TTYPE_ASSIGN) ||
		   	    (pltNext->ucType == TTYPE_ASSIGN_CMP))
					sRC = LocalAssign (&pltCurr, sFunc);
				else {
			   	ptTemp = pltCurr->ptToken;
			   	if (ptTemp->ucType == TTYPE_LABEL) {
						ptTemp = AddToken (0, TTYPE_NUM, 0, 0, &sRC);
						if (sRC == 0) {
						  	sRC = Eval (&pltCurr, ptTemp, sFunc);
							DeleteToken (ptTemp);
						}
					} else
						sRC = LocalLaunch (&pltCurr, sFunc);
				}
			} else
			  	SkipLine (&pltCurr);
		   *pltStart = pltCurr; 
	   	break;

   	case TTYPE_ENDBLK:
   	case TTYPE_THENELSE:
   	case TTYPE_LSTATEMENT:
  			pstTokIndirect = (PSTATEMENTTOK)pltCurr->ptToken;
			sRC = (*pstTokIndirect->Fxn)(pltStart, sFunc);
   		break;

		// Added in WinCmd 1.2 to allow statements in libs
   	case TTYPE_FSTATEMENT:
  			ptTemp = pltCurr->ptToken;
			// Modified in WinCmd 1.3 to pass proper parameters
			sRC = ((LPLIBFUNC)*(lpfnLibFunc[HIWORD(ptTemp->dwData)]))
			                (pltStart, (PTOKEN) sFunc, LOWORD (ptTemp->dwData),
			                 lpfnTokenFuncCallback, hMain);
   		break;

		case TTYPE_LIBFUNC:
		case TTYPE_LOCALFUNC:
			if (sFunc == FUNC_EXECUTE) {
				ptTemp = AddToken (0, TTYPE_NUM, 0, 0, &sRC);
				if (sRC == 0) {
				  	sRC = Eval (&pltCurr, ptTemp, sFunc);
					DeleteToken (ptTemp);
					if (pltCurr->ucType == TTYPE_EOL)
				   	(PBYTE)pltCurr += (WORD) pltCurr->ucLen;
				}
			} else
			  	SkipLine (&pltCurr);
		   *pltStart = pltCurr; 
			break;

   	case TTYPE_EOL:
		   (PBYTE)pltCurr += (WORD) pltCurr->ucLen;
		   *pltStart = pltCurr; 
   	   break;

   	case TTYPE_EOF:
		   *pltStart = pltCurr; 
   	   sRC = RC_END_OF_PROGRAM;
			break;
   		
		default:
			sRC = ERR_SYNTAX;
			break;
	}
	return sRC;
}

//------------------------------------------------------------
// LocalAssign - Assign a value to a token
//------------------------------------------------------------ 
INT LocalAssign (LPLINETOKEN *pltStart, INT sFunc){
	LPLINETOKEN	pltNext;
	INT			sRC = 0;
	PTOKEN		ptTemp, ptToken;

	if (sFunc & FUNC_SKIP) {
      SkipLine (pltStart);
		return 0;
	}
	pltNext = *pltStart;
	ptToken = (*pltStart)->ptToken;
   (PBYTE) pltNext += pltNext->ucLen;
  	(PBYTE) pltNext += pltNext->ucLen;
  	
	ptTemp = AddToken (0, TTYPE_NUM, 0, 0, &sRC);
	if (sRC == 0) {
	  	sRC = Eval (&pltNext, ptTemp, sFunc);
		if ((sRC == 0) && (sFunc & FUNC_EXECUTE)) {
			ptToken = (*pltStart)->ptToken;
			if (ptTemp->ucType == TTYPE_NUM) 
				sRC = SetToken (ptToken, ptTemp->ucType, ptTemp->wData, 
				                GetTokenVal (ptTemp));
			else	                
				sRC = SetToken (ptToken, ptTemp->ucType, ptTemp->wData, 
				                (DWORD)(LPSTR)GetTokenString(ptTemp));
		}
		DeleteToken (ptTemp);
		if ((sRC == 0) && (pltNext->ucType != TTYPE_EOL))
			sRC = ERR_EXTRA_CHARS;
		(PBYTE)pltNext += (WORD) pltNext->ucLen;
	}	
	*pltStart = pltNext; 
	return sRC;
}		
//------------------------------------------------------------
// LocalLaunch - Launch a program
//------------------------------------------------------------ 
INT LocalLaunch (LPLINETOKEN *pltStart, INT sFunc){
   char	szProg[MAXLINELEN];
   INT	sRC;
	LONG	lResult;

	if (sFunc & FUNC_SKIP) {
      SkipLine (pltStart);
		return 0;
	}
	if ((*pltStart)->ucType == TTYPE_LINE)
		*(PWORD)pltStart += (*pltStart)->ucLen;

	sRC = ResolveLine (pltStart, szProg, sizeof (szProg) - 5);
	if (sRC) return sRC;
	
	lResult = LaunchString (szProg, 0);
	sRC = HIWORD (lResult);
	*(PWORD)pltStart += (*pltStart)->ucLen;
   return sRC;
}
//===========================================================
// NotifyCallback - Callback function for toolhelp 
// RegisterNotify function.  Allows tracking of app 
// termination.
//===========================================================
BOOL CALLBACK THNotifyCallback (WORD wID, DWORD dwData) {
	HANDLE hTask;
	TASKENTRY te;
	PTOKEN pToken;
	
	if (wID == NFY_EXITTASK) {
		hTask = GetCurrentTask();		
		te.dwSize = sizeof(TASKENTRY);
		((THTFH)*lpfnTaskFindHandle)(&te, hTask);

      if (te.hInst == hHoldTask) {
			//Clear return code token
			pToken = FindToken ("RETURNCODE", 10);
			if (pToken)
				SetToken (pToken, TTYPE_NUM, 0, LOWORD (dwData));
			
			PostMessage (hMain, MYMSG_RESTART, TRUE, 0);
		}	
	}
	return FALSE;
}	
//------------------------------------------------------------
// LocalLaunchandWait - Launch a program then stop until 
// launched program complete.  Also allows custom show_window
// parameters.
//------------------------------------------------------------ 
INT LocalLaunchandWait (LPLINETOKEN *pltStart, INT sFunc){
   char	szProg[MAXLINELEN];
	PTOKEN pToken;
   INT	sRC = 0;
	UINT	usStartCode;
	LONG	lResult;

	if (sFunc & FUNC_SKIP) {
      SkipLine (pltStart);
		return 0;
	}
	if ((*pltStart)->ucType == TTYPE_LINE)
		*(PWORD)pltStart += (*pltStart)->ucLen;
	//Move past LAW token
	*(PWORD)pltStart += (*pltStart)->ucLen;
	//Evaluate launch code
	usStartCode = (UINT)Eval2Num (pltStart, sFunc, &sRC);
	if (sRC) return sRC;

	if ((usStartCode == 0) || ((usStartCode & 0xff) < 5))
		sRC = ERR_BADLAUNCHCODE;

	//Covert line to text
	sRC = ResolveLine (pltStart, szProg, sizeof (szProg) - 5);
	if (sRC) return sRC;		

	//Load ToolHelp dll to get notify function.
	if (lpfnNotifyRegister == 0) {
		hToolHelp = LoadLibrary ("TOOLHELP.DLL");
		if (hToolHelp < HINSTANCE_ERROR)
			return ERR_NOTOOLHELP;
		lpfnNotifyRegister = GetProcAddress (hToolHelp, "NotifyRegister");
		lpfnNotifyUnRegister = GetProcAddress (hToolHelp, "NotifyUnRegister");
		lpfnTaskFindHandle = GetProcAddress (hToolHelp, "TaskFindHandle");
		if (!(lpfnNotifyRegister && lpfnNotifyUnRegister && lpfnTaskFindHandle))
			return ERR_NOTOOLHELP;
		lpfnTHCallBack = MakeProcInstance (THNotifyCallback, hInst);
	}		
	if (hHoldTask) 
		((THUREGPROC)*lpfnNotifyUnRegister)(NULL);
	//Clear return code token
	pToken = FindToken ("RETURNCODE", 10);
	if (pToken)
		SetToken (pToken, TTYPE_NUM, 0, 0x10000);

	lResult = LaunchString (szProg, usStartCode & 0xff);
	sRC = HIWORD (lResult);
	if ((sRC == 0) && (usStartCode & LC_WAIT)) {
		hHoldTask = LOWORD (lResult);
		if (((THREGPROC)*lpfnNotifyRegister)(NULL, lpfnTHCallBack, NF_NORMAL)) {
			SetRunning (FALSE);
			hTimer2 = SetTimer (hMain, 2, 1000, 0);
			if (hTimer2 == 0)
				sRC = ERR_NO_TIMERS;
		}
	}	
	*(PWORD)pltStart += (*pltStart)->ucLen;
   return sRC;
}

//------------------------------------------------------------
// LocalOnDrop - ONDROP statement
//------------------------------------------------------------ 
INT LocalOnDrop (LPLINETOKEN *pltStart, INT sFunc) {
	INT sRC;
	HWND hWnd = hMain;
	PTOKEN pToken;

	if ((*pltStart)->ucType == TTYPE_LINE)
		*(LPWORD)pltStart += (*pltStart)->ucLen;
	//Move past ONDROP statement
	*(LPWORD)pltStart += (*pltStart)->ucLen;

	if (!(sFunc & FUNC_SKIP)) {
		pToken = FindToken ("_DROPPTR", 8);
		if (pToken) {
			if (((*pltStart)->ucType == TTYPE_EOL) ||
			    ((*pltStart)->ucType == TTYPE_STATEMENTOFF)) {
				DragAcceptFiles (hWnd, FALSE);
				pToken->dwData = 0;
			} else {
				DragAcceptFiles (hWnd, TRUE);
				pToken->dwData = (DWORD)*pltStart;
			}
		}	
	}
	sRC = Execute (pltStart, FUNC_SKIP);
	if ((*pltStart)->ucType == TTYPE_EOL)
	   *(LPWORD)pltStart += (*pltStart)->ucLen;

	return sRC;
}
//------------------------------------------------------------
// While - While loop statement
//------------------------------------------------------------ 
INT LocalWHILE (LPLINETOKEN *pltStart, INT sFunc){
	LPLINETOKEN	pltNext, pltSave;
	INT			sRC = 0;

	pltNext = *pltStart;
	if (pltNext->ucType == TTYPE_LINE)
	   (PBYTE) pltNext += pltNext->ucLen;
   (PBYTE) pltNext += pltNext->ucLen;
	
	if (!Eval2Num (&pltNext, sFunc, &sRC))
		sFunc = FUNC_SKIP;
	if (sRC) return sRC;

	pltSave = pltNext;	
	if (pltNext->ucType == TTYPE_EOL) {
	   (PBYTE) pltNext += pltNext->ucLen;
		sRC = ExecuteLine (&pltNext, sFunc);
	} else {
		sRC = Execute (&pltNext, sFunc);
		if (pltNext->ucType == TTYPE_EOL)
		   (PBYTE) pltNext += pltNext->ucLen;
	}
	if (sRC == RC_END_OF_LOOP) {
	   sRC = 0;
	   pltNext = pltSave;
		if (pltNext->ucType == TTYPE_EOL) {
		   (PBYTE) pltNext += pltNext->ucLen;
			sRC = ExecuteLine (&pltNext, FUNC_SKIP);
		} else {
			sRC = Execute (&pltNext, FUNC_SKIP);
			if (pltNext->ucType == TTYPE_EOL)
			   (PBYTE) pltNext += pltNext->ucLen;
		}
		*pltStart = pltNext;
	}	
	if ((sFunc & FUNC_SKIP) || sRC)
		*pltStart = pltNext;

	return sRC;
}

//------------------------------------------------------------
// IF - If statement
//------------------------------------------------------------ 
INT LocalIF (LPLINETOKEN *pltStart, INT sFunc) {
	LPLINETOKEN	pltNext;
	LONG			lResult;
	INT			sRC = 0;

	pltNext = *pltStart;
	if (pltNext->ucType == TTYPE_LINE)
	   (PBYTE) pltNext += pltNext->ucLen;
   (PBYTE) pltNext += pltNext->ucLen;
	
	lResult = Eval2Num (&pltNext, sFunc, &sRC);
	if (sRC) return sRC;

	if (pltNext->ucType != TTYPE_EOL) {  // If <then> clause on same line as IF
	   if (lResult) {
	      if (sFunc == FUNC_EXECUTE) {
	      	sRC = Execute (&pltNext, sFunc);
	      	if (pltNext->ucType == TTYPE_EOL) 
   				(PBYTE) pltNext += pltNext->ucLen;
	      }
	      if (GetStatementType (pltNext) == TTYPE_THENELSE)
				sRC = ExecuteLine (&pltNext, FUNC_SKIP);
		} else 
      	sRC = Execute (&pltNext, FUNC_SKIP);

	} else { // If <then> clause on next line
		(PBYTE) pltNext += pltNext->ucLen;
	   if (lResult) {
			sRC = ExecuteLine (&pltNext, sFunc);
	      if (GetStatementType (pltNext) == TTYPE_THENELSE) {
				sRC = ExecuteLine (&pltNext, FUNC_SKIP);
			}	
		} else 
			sRC = ExecuteLine (&pltNext, FUNC_SKIP);
	}
	*pltStart = pltNext;
	return sRC;
}

//------------------------------------------------------------
// Then/Else Statements
//------------------------------------------------------------ 
INT LocalThenElse (LPLINETOKEN *pltStart, INT sFunc) {
	LPLINETOKEN	pltNext;
	INT         sRC = 0;

	pltNext = *pltStart;
	if (pltNext->ucType == TTYPE_LINE)
	   (PBYTE) pltNext += pltNext->ucLen;
   (PBYTE) pltNext += pltNext->ucLen;
	
	if (pltNext->ucType != TTYPE_EOL)      // More on same line
	   sRC = Execute (&pltNext, sFunc);
	else {                                 // Execute Next line
  		(PBYTE) pltNext += pltNext->ucLen;
		sRC = ExecuteLine (&pltNext, sFunc);
   }
	*pltStart = pltNext;
	return sRC;
}

//------------------------------------------------------------
// LocalDO - DO statement
//------------------------------------------------------------ 
INT LocalDO (LPLINETOKEN *pltStart, INT sFunc) {
	LPLINETOKEN	pltNext;
	INT			sRC = 0;

	pltNext = *pltStart;
	if (pltNext->ucType == TTYPE_LINE)
	   (PBYTE) pltNext += pltNext->ucLen;
   (PBYTE) pltNext += pltNext->ucLen;

	if (pltNext->ucType != TTYPE_EOL) 
		return  ERR_EXTRA_CHARS;

	(PBYTE) pltNext += pltNext->ucLen;

	while (sRC == 0) 
		sRC = ExecuteLine (&pltNext, sFunc);

	*pltStart = pltNext;
	if (sRC != RC_END_OF_BLOCK) 
		return sRC;
	return 0;
}

//------------------------------------------------------------
// LocalEND - END statement ends a block
//------------------------------------------------------------ 
INT LocalEND (LPLINETOKEN *pltStart, INT sFunc) {

	if ((*pltStart)->ucType == TTYPE_LINE)
		*(PWORD)pltStart += (*pltStart)->ucLen;
	*(PWORD)pltStart += (*pltStart)->ucLen;

	if ((*pltStart)->ucType != TTYPE_EOL) 
		return  ERR_EXTRA_CHARS;

	*(PWORD)pltStart += (*pltStart)->ucLen;
	return RC_END_OF_BLOCK;
}

//------------------------------------------------------------
// LocalRETURN - RETURN statement returns from a sub or
// function.
//------------------------------------------------------------ 
INT LocalRETURN (LPLINETOKEN *pltStart, INT sFunc) {
   INT		sRC = 0;
   PTOKEN	ptRetVal;

	if (sFunc & FUNC_SKIP) {
      SkipLine (pltStart);
		return 0;
	}
	if ((*pltStart)->ucType == TTYPE_LINE)
		*(PWORD)pltStart += (*pltStart)->ucLen;
	*(PWORD)pltStart += (*pltStart)->ucLen;

	
	ptRetVal = FindTokenByType (TTYPE_RETVAL);
	ptRetVal->ucType = TTYPE_NUM;

	if ((*pltStart)->ucType == TTYPE_EOL) 
		return RC_END_OF_SUB;

	if (ptRetVal != 0) {
		sRC = Eval (pltStart, ptRetVal, sFunc);
		if (sRC) 
			return sRC;
	}
	if ((*pltStart)->ucType == TTYPE_EOL) 
		*(PWORD)pltStart += (*pltStart)->ucLen;
	return RC_END_OF_SUB;
}
//------------------------------------------------------------
// LocalLEAVE - LEAVE statement exits from a loop
//------------------------------------------------------------ 
INT LocalLEAVE (LPLINETOKEN *pltStart, INT sFunc) {

	if (sFunc & FUNC_SKIP) {
      SkipLine (pltStart);
		return 0;
	}
	if ((*pltStart)->ucType == TTYPE_LINE)
		*(PWORD)pltStart += (*pltStart)->ucLen;
	*(PWORD)pltStart += (*pltStart)->ucLen;

	if ((*pltStart)->ucType != TTYPE_EOL) 
		return ERR_EXTRA_CHARS;
	else
		return RC_END_OF_LOOP;
}
//------------------------------------------------------------
// LocalCMT2EOL- Comment statment, line ignored
// function.
//------------------------------------------------------------ 
INT LocalCMT2EOL (LPLINETOKEN *pltStart, INT sFunc) {
	SkipLine (pltStart);
	return 0;
}
//------------------------------------------------------------
// LocalEXIT - EXIT statement terminates program
//------------------------------------------------------------ 
INT LocalEXIT (LPLINETOKEN *pltStart, INT sFunc) {
   INT	sRC = 0;

	if (sFunc & FUNC_SKIP) {
      SkipLine (pltStart);
		return 0;
	}
	if ((*pltStart)->ucType == TTYPE_LINE)
		*(PWORD)pltStart += (*pltStart)->ucLen;
	*(PWORD)pltStart += (*pltStart)->ucLen;

	if ((*pltStart)->ucType == TTYPE_EOL) 
		return RC_END_OF_PROGRAM;
	
	sFinalRC = (INT) Eval2Num (pltStart, sFunc, &sRC);
	if (sRC) 
		return sRC;
	else
		return RC_END_OF_PROGRAM;
}

//------------------------------------------------------------
// LocalSTOP - STOP statement halts program
//------------------------------------------------------------ 
INT LocalSTOP (LPLINETOKEN *pltStart, INT sFunc) {

	if (sFunc & FUNC_SKIP) {
      SkipLine (pltStart);
		return 0;
	}
	if ((*pltStart)->ucType == TTYPE_LINE)
		*(PWORD)pltStart += (*pltStart)->ucLen;
	*(PWORD)pltStart += (*pltStart)->ucLen;

	SetRunning (FALSE);
	if ((*pltStart)->ucType != TTYPE_EOL) 
		return ERR_EXTRA_CHARS;
	//Skip past EOL token		
	*(PWORD)pltStart += (*pltStart)->ucLen;
	return 0;
}
//------------------------------------------------------------
// LocalSAY - SAY statement
//------------------------------------------------------------ 
INT LocalSAY (LPLINETOKEN *pltStart, INT sFunc) {
   char	szOutput[MAXLINELEN];
   INT	sRC;

	if (sFunc & FUNC_SKIP) {
      SkipLine (pltStart);
		return 0;
	}
	if ((*pltStart)->ucType == TTYPE_LINE)
		*(PWORD)pltStart += (*pltStart)->ucLen;
	*(PWORD)pltStart += (*pltStart)->ucLen;

	sRC = ResolveLine (pltStart, szOutput, sizeof (szOutput));
	if (sRC) return sRC;
	WriteOutput (hMain, szOutput);	
	*(PWORD)pltStart += (*pltStart)->ucLen;
	return sRC;
}
//------------------------------------------------------------
// LocalChDir - ChDir changes the current directory
//------------------------------------------------------------ 
INT LocalChDir (LPLINETOKEN *pltStart, INT sFunc) {
   char	szOutput[MAXLINELEN];
   INT	sRC;

	if (sFunc & FUNC_SKIP) {
      SkipLine (pltStart);
		return 0;
	}
	if ((*pltStart)->ucType == TTYPE_LINE)
		*(PWORD)pltStart += (*pltStart)->ucLen;
	*(PWORD)pltStart += (*pltStart)->ucLen;

	if ((*pltStart)->ucType != TTYPE_EOL) {
		sRC = ResolveLine (pltStart, szOutput, sizeof (szOutput));
		if (sRC) return sRC;
		if (strlen (szOutput) < 128) 
			sRC = chdir (_fullpath(&szOutput[128], szOutput, sizeof (szOutput)-128));
		else						
			sRC = chdir(szOutput);
		if (sRC != 0)
		   sRC = ERR_PATH_NOT_FOUND;
	} else {
	   getcwd (szOutput, sizeof (szOutput));
		WriteOutput (hMain, szOutput);
		sRC = 0;	   
	}  
	*(PWORD)pltStart += (*pltStart)->ucLen;
	return sRC;
}
//------------------------------------------------------------
// LocalCLEAROUTPPPUT - CLEAROUTPUT statement clears the output
// listbox
//------------------------------------------------------------ 
INT LocalCLEAROUTPUT (LPLINETOKEN *pltStart, INT sFunc) {

	if (sFunc & FUNC_SKIP) {
      SkipLine (pltStart);
		return 0;
	}
	if ((*pltStart)->ucType == TTYPE_LINE)
		*(PWORD)pltStart += (*pltStart)->ucLen;
	*(PWORD)pltStart += (*pltStart)->ucLen;

	SendDlgItemMessage (hMain, IDD_OUTLIST, LB_RESETCONTENT, 0, 0);

	if ((*pltStart)->ucType != TTYPE_EOL) 
		return ERR_EXTRA_CHARS;
	return 0;
}
//------------------------------------------------------------
// Operators 
//------------------------------------------------------------ 
//------------------------------------------------------------
// LocalOpAdd - Adds two numbers
//------------------------------------------------------------ 
INT LocalOpAdd (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {
   LONG	lTemp, lSrc, lDest;
   INT	sRC = 0;

	if (ptDest->ucType == TTYPE_STR) {
		sRC = AppendString (ptDest, ptSrc);
	} else {
		lSrc = GetTokenVal (ptSrc);
		lDest = GetTokenVal (ptDest);
		lTemp = lSrc + lDest;
		if (!((lDest ^ lSrc)  & 0x80000000) &&
		     ((lDest ^ lTemp) & 0x80000000)) 
			sRC = ERR_OVERFLOW;
		else
			sRC = SetToken (ptDest, TTYPE_NUM, 0, lTemp);
	}	
  	return sRC;
}  	
//------------------------------------------------------------
// LocalOpSub - Subtracts two numbers
//------------------------------------------------------------ 
INT LocalOpSub (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {
   LONG	lTemp, lSrc, lDest;
   INT	sRC = 0;

	lSrc = GetTokenVal (ptSrc);
	lDest = GetTokenVal (ptDest);
	lTemp = lDest - lSrc;
	if (((lDest ^ lSrc) & 0x80000000) &&
	    ((lDest ^ lTemp) & 0x80000000)) 
		sRC = ERR_OVERFLOW;
	else
		sRC = SetToken (ptDest, TTYPE_NUM, 0, lTemp);
  	return sRC;
}
//------------------------------------------------------------
// LocalOpMul - Multiplies two numbers
//------------------------------------------------------------ 
INT LocalOpMul (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {
   LONG	lTemp, lSrc, lDest;
   INT	sRC = 0;

	lSrc = GetTokenVal (ptSrc);
	lDest = GetTokenVal (ptDest);
	lTemp = lDest * lSrc;
	if (((lDest ^ lSrc) & 0x80000000) && (lTemp >= 0) ||
	    !((lDest ^ lSrc) & 0x80000000) && (lTemp < 0)) 
		sRC = ERR_OVERFLOW;
	else
		sRC = SetToken (ptDest, TTYPE_NUM, 0, lTemp);
  	return sRC;
}  	
//------------------------------------------------------------
// LocalOpDiv - Divides two numbers
//------------------------------------------------------------ 
INT LocalOpDiv (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {
   LONG lTemp;

   lTemp = GetTokenVal (ptSrc);
	if (lTemp == 0) return ERR_DIVIDE_BY_ZERO;
	return SetToken (ptDest, TTYPE_NUM, 0, GetTokenVal(ptDest) / lTemp); 
}  	
//------------------------------------------------------------
// LocalOpMod - Returns the remainder after dividing two numbers
//------------------------------------------------------------ 
INT LocalOpMod (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {
   LONG lTemp;

   lTemp = GetTokenVal (ptSrc);
	if (lTemp == 0) return ERR_DIVIDE_BY_ZERO;
	return SetToken (ptDest, TTYPE_NUM, 0, GetTokenVal(ptDest) % lTemp); 
}  	
//------------------------------------------------------------
// LocalOpLThan - Less than comparision
//------------------------------------------------------------ 
INT LocalOpLThan (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {
   LONG	lSrc, lDest;

	lSrc = GetTokenVal (ptSrc);
	lDest = GetTokenVal (ptDest);
	if (lDest < lSrc)
		return SetToken (ptDest, TTYPE_NUM, 0, MYTRUE); 
	else
		return SetToken (ptDest, TTYPE_NUM, 0, MYFALSE); 
}  	
//------------------------------------------------------------
// LocalOpLEThan - Less or equal comparision
//------------------------------------------------------------ 
INT LocalOpLEThan (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {
   LONG	lSrc, lDest;

	lSrc = GetTokenVal (ptSrc);
	lDest = GetTokenVal (ptDest);
	if (lDest <= lSrc)
		return SetToken (ptDest, TTYPE_NUM, 0, MYTRUE); 
	else
		return SetToken (ptDest, TTYPE_NUM, 0, MYFALSE); 
}  	
//------------------------------------------------------------
// LocalOpGThan - Greater than comparision
//------------------------------------------------------------ 
INT LocalOpGThan (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {
   LONG	lSrc, lDest;

	lSrc = GetTokenVal (ptSrc);
	lDest = GetTokenVal (ptDest);
	if (lDest > lSrc)
		return SetToken (ptDest, TTYPE_NUM, 0, MYTRUE); 
	else
		return SetToken (ptDest, TTYPE_NUM, 0, MYFALSE); 
}  	
//------------------------------------------------------------
// LocalOpGEThan - Greater or equal comparision
//------------------------------------------------------------ 
INT LocalOpGEThan (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {
   LONG	lSrc, lDest;

	lSrc = GetTokenVal (ptSrc);
	lDest = GetTokenVal (ptDest);
	if (lDest >= lSrc)
		return SetToken (ptDest, TTYPE_NUM, 0, MYTRUE); 
	else
		return SetToken (ptDest, TTYPE_NUM, 0, MYFALSE); 
}  	
//------------------------------------------------------------
// LocalOpNotEqual - Not equal comparison
//------------------------------------------------------------ 
INT LocalOpNotEqual (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {
   INT	sRC;
   sRC = LocalOpCmp (pltLine, ptDest, ptSrc);
   if (sRC == 0)
		if (GetTokenVal (ptDest) == MYFALSE)
			return SetToken (ptDest, TTYPE_NUM, 0, MYTRUE); 
		else
			return SetToken (ptDest, TTYPE_NUM, 0, MYFALSE); 
  	return sRC;
}  	
//------------------------------------------------------------
// LocalOpCmp - Equal comparison
//------------------------------------------------------------ 
INT LocalOpCmp (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {

	if (ptDest->ucType == TTYPE_NUM) {
	   if (GetTokenVal (ptDest) == GetTokenVal(ptSrc))
			return SetToken (ptDest, TTYPE_NUM, 0, MYTRUE); 
		else
			return SetToken (ptDest, TTYPE_NUM, 0, MYFALSE); 

	} else if (ptDest->ucType == TTYPE_STR) {
        if (lstrcmp ((LPSTR)GetTokenString(ptDest),(LPSTR)GetTokenString(ptSrc)) == 0)
			return SetToken (ptDest, TTYPE_NUM, 0, MYTRUE); 
		else
			return SetToken (ptDest, TTYPE_NUM, 0, MYFALSE); 

	} else if ((ptDest->ucType == ptSrc->ucType) && 
	           (ptDest->wData == ptSrc->wData) &&
	           (ptDest->dwData == ptSrc->dwData)) {
		return SetToken (ptDest, TTYPE_NUM, 0, MYTRUE); 
	}
	return SetToken (ptDest, TTYPE_NUM, 0, MYFALSE); 
}  	
//------------------------------------------------------------
// LocalOpAND - Logical AND
//------------------------------------------------------------ 
INT LocalOpAND (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {

   if (GetTokenVal (ptSrc) && GetTokenVal (ptDest))
		SetToken (ptDest, TTYPE_NUM, 0, MYTRUE);
	else
		SetToken (ptDest, TTYPE_NUM, 0, MYFALSE);
  	return 0;
}  	
//------------------------------------------------------------
// LocalOpANDBits - Bitwise AND
//------------------------------------------------------------ 
INT LocalOpANDBits (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {
	return SetToken (ptDest, TTYPE_NUM, 0, GetTokenVal (ptSrc) & 
	                                       GetTokenVal (ptDest));
}  	
//------------------------------------------------------------
// LocalOpOR - Logical OR
//------------------------------------------------------------ 
INT LocalOpOR (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {

   if (GetTokenVal (ptSrc) || GetTokenVal (ptDest))
		SetToken (ptDest, TTYPE_NUM, 0, MYTRUE);
	else
		SetToken (ptDest, TTYPE_NUM, 0, MYFALSE);
  	return 0;
}  	
//------------------------------------------------------------
// LocalOpORBits - Bitwise OR
//------------------------------------------------------------ 
INT LocalOpORBits (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {
	return SetToken (ptDest, TTYPE_NUM, 0, GetTokenVal (ptSrc) | 
	                                       GetTokenVal (ptDest));
}  	
//------------------------------------------------------------
// LocalOpXOR - Logical XOR
//------------------------------------------------------------ 
INT LocalOpXOR (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {

   if ( (GetTokenVal (ptSrc) && GetTokenVal (ptDest)) ||
       !(GetTokenVal (ptSrc) && GetTokenVal (ptDest)))
		SetToken (ptDest, TTYPE_NUM, 0, MYTRUE);
	else
		SetToken (ptDest, TTYPE_NUM, 0, MYFALSE);
  	return 0;
}  	
//------------------------------------------------------------
// LocalOpXORBits - Bitwise XOR
//------------------------------------------------------------ 
INT LocalOpXORBits (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {
	return SetToken (ptDest, TTYPE_NUM, 0, GetTokenVal (ptSrc) ^ 
	                                       GetTokenVal (ptDest));
}  	
//------------------------------------------------------------
// LocalOpNOT - Logical NOT
//------------------------------------------------------------ 
INT LocalOpNOT (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {

   if (GetTokenVal (ptSrc))
		SetToken (ptDest, TTYPE_NUM, 0, MYFALSE);
	else
		SetToken (ptDest, TTYPE_NUM, 0, MYTRUE);
  	return 0;
}  	
//------------------------------------------------------------
// LocalOpNOTBits - Bitwise NOT
//------------------------------------------------------------ 
INT LocalOpNOTBits (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {
	return SetToken (ptDest, TTYPE_NUM, 0, ~GetTokenVal (ptSrc));
}  	
//------------------------------------------------------------
// LocalOpOComma - Pauses current evaluation
//------------------------------------------------------------ 
INT LocalComma (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {
  	return 0;
}  	
//------------------------------------------------------------
// LocalOpOParen - Processes open paren character
//------------------------------------------------------------ 
INT LocalOpOPren (LPLINETOKEN *pltLine, PTOKEN ptDest) {
   INT	sRC;

	sRC = Eval (pltLine, ptDest, FUNC_EXECUTE);
	if (sRC) return sRC;
	if ((*pltLine)->ucType != TTYPE_CPAREN) 
      return ERR_NO_CLOSE_PAREN;

	sRC = (*pltLine)->ucLen;
	*(PWORD)pltLine += sRC;
	sRC = 0;

	return sRC;			
}
INT LocalOpCPren (LPLINETOKEN *pltLine, PTOKEN ptDest, PTOKEN ptSrc) {
  	return ERR_NO_OPEN_PAREN;
}  	
//------------------------------------------------------------
// Predefined functions
//------------------------------------------------------------ 
//------------------------------------------------------------
// LocalLENGTH - Predefined func that returns the length of a string
//------------------------------------------------------------ 
INT LocalLENGTH (LPLINETOKEN *pltLine, PTOKEN ptDest) {
   PTOKEN	ptArg;
  
	if (ptDest->wData < 2) 
		return ERR_TOO_FEW_PARMS;

	ptArg = (PTOKEN) ptDest->dwData;   
   ptDest->ucType = TTYPE_NUM;
	return SetToken (ptDest, TTYPE_NUM, 0, 
	                 strlen (GetTokenString (ptArg-1)));
}  	
//------------------------------------------------------------
// LocalUPPER - Predefined func that converts a string to
// upper case.
//------------------------------------------------------------ 
INT LocalUPPER (LPLINETOKEN *pltLine, PTOKEN ptDest) {
   PTOKEN	ptArg;

	if (ptDest->wData < 2) 
		return ERR_TOO_FEW_PARMS;
   
	ptArg = (PTOKEN) ptDest->dwData;   
   ptDest->ucType = TTYPE_NUM;
	return SetToken (ptDest, TTYPE_STR, 0, 
		           (DWORD)(LPSTR)strupr(GetTokenString(ptArg-1)));
}  	
//------------------------------------------------------------
// LocalSUBSTR - Predefined func that returns a portion of a 
// string.
// Arg 1 - Input string
// Arg 2 - Starting character
// Arg 3 - Length
//------------------------------------------------------------ 
INT LocalSUBSTR (LPLINETOKEN *pltLine, PTOKEN ptDest) {
   INT		sRC = 0;
   UINT		wStart = 0, wLen = 0xffff, wStrLen;
   char		chChar;
   char		*pszSrc;
   PTOKEN	ptArg;

	if (ptDest->wData < 2) 
		return ERR_TOO_FEW_PARMS;
   
   ptArg = (PTOKEN) ptDest->dwData;   
	if (ptDest->wData > 3)
	   wLen = (WORD) GetTokenVal (ptArg-3);

	if (ptDest->wData > 2)
	   wStart = (WORD) GetTokenVal (ptArg-2);
	
	pszSrc = GetTokenString (ptArg-1);	
	wStrLen = strlen (pszSrc);
	if (wStart > wStrLen)
		sRC = SetToken (ptDest, TTYPE_STR, 0, (DWORD)(LPSTR)"");
	else {
	   pszSrc += wStart;
		wStrLen = strlen (pszSrc);
		if (wLen < wStrLen) {
		   chChar = *(pszSrc+wLen);
		   *(pszSrc+wLen) = 0;
		}
		sRC = SetToken (ptDest, TTYPE_STR, 0, (DWORD)(LPSTR)pszSrc);
		if (wLen < wStrLen) 
		   *(pszSrc+wLen) = chChar;
	}
  	return sRC;
}  	
//------------------------------------------------------------
// LocalARG - Predefined func that returns an a calling arg.
//------------------------------------------------------------ 
INT LocalARG (LPLINETOKEN *pltLine, PTOKEN ptDest) {
   INT		sRC;
   PTOKEN	ptArg, ptRetVal;

   ptDest->ucType = TTYPE_NUM;
	ptRetVal = FindTokenByType (TTYPE_RETVAL);
	if (ptRetVal == 0)
	   return ERR_NOARGS;
	if (ptDest->wData == 1) {
		sRC = SetToken (ptDest, TTYPE_NUM, 0, (DWORD)ptRetVal->wData-1);
	} else {
	   ptArg = (PTOKEN) ptDest->dwData;
		if (ptRetVal->wData <= (UINT) GetTokenVal (ptArg-1)) {
			sRC = SetToken (ptDest, TTYPE_STR, 0, (DWORD)(LPSTR)szNull);
		} else {		   
			ptRetVal = (PTOKEN)ptRetVal->dwData;
			ptRetVal -= GetTokenVal (ptArg-1);
			sRC = SetToken (ptDest, ptRetVal->ucType, ptRetVal->wData, ptRetVal->dwData);
		}
	}			   
  	return sRC;
}  	
//------------------------------------------------------------
// LocalWAIT - Predefined func that stops execution for a 
// number of milliseconds.
// Arg 1 - Number of milliseconds to wait
//------------------------------------------------------------ 
INT LocalWAIT (LPLINETOKEN *pltLine, PTOKEN ptDest) {
	PTOKEN	ptArg;
	UINT		wDelay;

	if (ptDest->wData < 2) 
		return ERR_TOO_FEW_PARMS;

	ptArg = (PTOKEN) ptDest->dwData;
	if (ptDest->wData < 2) 
		return SetToken (ptDest, TTYPE_NUM, 0, (DWORD)MYFALSE);
	
	wDelay = (UINT) GetTokenVal (ptArg-1);
	if (wDelay == 0)
		return SetToken (ptDest, TTYPE_NUM, 0, (DWORD)MYFALSE);
	if (GetTokenVal (ptArg-1) > 0xffff)
	   wDelay = 0xffff;

	hWaitTimer = SetTimer (hMain, WAITTIMER_ID, wDelay, 0);
	if (hWaitTimer == 0)
		return ERR_NO_TIMERS;
	SetRunning (FALSE);
	return SetToken (ptDest, TTYPE_NUM, 0, (DWORD)MYTRUE);
}  	
//------------------------------------------------------------
// LocalFindLib - Predefined func that returns lib loaded 
// Arg 1 - Name of lib to look for
//------------------------------------------------------------ 
INT LocalFindLib (LPLINETOKEN *pltLine, PTOKEN ptDest) {
	PTOKEN	ptArg;
	FARPROC	lpfnLibCall;
	INT		i, sVer;
	BOOL		bFound = FALSE;
	char		szLibName[MAXSTRLEN];

	if (ptDest->wData < 2) 
		return ERR_TOO_FEW_PARMS;

	ptArg = (PTOKEN) ptDest->dwData;
	strupr (GetTokenString (ptArg-1));
	for (i = 0; i < sLibCount; i++) {
		lpfnLibCall = GetProcAddress (hLib[i], "WCGetVersion");
		((LPLIBGETVER)*lpfnLibCall) ((LPINT) &sVer, (LPSTR) szLibName,
		                             sizeof (szLibName));
		strupr (szLibName);
		if (strcmp (szLibName, GetTokenString (ptArg-1)) == 0) {
		   bFound = TRUE;
		   break;
		}		
	}			   
	if (bFound)
		return SetToken (ptDest, TTYPE_NUM, 0, (DWORD) sVer);
	else
		return SetToken (ptDest, TTYPE_NUM, 0, 0);
}  	
//------------------------------------------------------------
// Predefined functions useful in Windows
//------------------------------------------------------------ 
INT MyGetWindowText (HWND hWnd, LPSTR lpszOut, UINT wLen) {
   return (INT) SendMessage (hWnd, WM_GETTEXT, wLen, (DWORD) lpszOut);
}   
//------------------------------------------------------------
// WinActivate - Predefined func that sets the input focus to
// a window
// Arg 1 - Window title text
//------------------------------------------------------------ 
INT WinActivate (LPLINETOKEN *pltLine, PTOKEN ptDest) {
	char		szWinText[MAXLINELEN];
	PTOKEN	ptArg;
	HWND		hWndTarg, hWndT2;
	
	if (ptDest->wData < 2) 
		return ERR_TOO_FEW_PARMS;

	ptArg = (PTOKEN) ptDest->dwData;   
	hWndTarg = FindWindow (0, GetTokenString(ptArg-1));
	if (hWndTarg != 0) {  
		hWndT2 = SetActiveWindow (hWndTarg);
		MyGetWindowText (hWndT2, szWinText, sizeof szWinText);
		return SetToken (ptDest, TTYPE_STR, 0,(DWORD)(LPSTR)szWinText);
	} else	
		return SetToken (ptDest, TTYPE_NUM, 0, (DWORD)MYFALSE);
}  	
//------------------------------------------------------------
// WinSendKeys - Predefined func that sends keys to a window
// Arg 1 - Key string
//------------------------------------------------------------ 
INT WinSendKeys (LPLINETOKEN *pltLine, PTOKEN ptDest) {
	PTOKEN	ptArg;
	char		*pszKeyString;
	INT		sRC;
	
	if (ptDest->wData > 1) {
		ptArg = (PTOKEN) ptDest->dwData;   
		pszKeyString = GetTokenString (ptArg-1);
		sRC = StuffKeyString (&pszKeyString);
		if (sRC) return sRC;
		return SetToken (ptDest, TTYPE_NUM, 0, (DWORD)MYTRUE);
	} else
		return ERR_TOO_FEW_PARMS;
}  	
//------------------------------------------------------------
// WinGetActWin - Predefined func that returns the text of
// the active window.
// No Args
//------------------------------------------------------------ 
INT WinGetActWin (LPLINETOKEN *pltLine, PTOKEN ptDest) {
	char		szWinText[MAXLINELEN];
	
	MyGetWindowText (GetActiveWindow(), szWinText, sizeof szWinText);
	return SetToken (ptDest, TTYPE_STR, 0, (DWORD)(LPSTR) szWinText);
}  	
//------------------------------------------------------------
// WinGetNextWin - Predefined func that returns the text of
// the next window in the window list.
// Arg 1 - Window title text
//------------------------------------------------------------ 
INT WinGetNextWin (LPLINETOKEN *pltLine, PTOKEN ptDest) {
	char		szWinText[MAXLINELEN];
	PTOKEN	ptArg;
	HWND		hWndTarg;
	
	if (ptDest->wData < 2) 
		return ERR_TOO_FEW_PARMS;

	ptArg = (PTOKEN) ptDest->dwData;   
	
	hWndTarg = FindWindow (0, GetTokenString(ptArg-1));
	hWndTarg = GetNextWindow(hWndTarg, GW_HWNDNEXT);
	if (hWndTarg == 0)
		return SetToken (ptDest, TTYPE_STR, 0, MYFALSE);
	MyGetWindowText (hWndTarg, szWinText, sizeof szWinText);
	return SetToken (ptDest, TTYPE_STR, 0, (DWORD)(LPSTR) szWinText);
}  	

//------------------------------------------------------------
// WinGetWinEXE - Predefined func that returns the name of the
// EXE that created a window
// Arg 1 - Window title text
//------------------------------------------------------------ 
INT WinGetWinEXE (LPLINETOKEN *pltLine, PTOKEN ptDest) {
	char		szWinText[MAXLINELEN];
	PTOKEN	ptArg;
	HWND		hWndTarg;
	
	ptArg = (PTOKEN) ptDest->dwData;
	if (ptDest->wData > 1) {
		hWndTarg = FindWindow (0, GetTokenString(ptArg-1));
		if (hWndTarg != 0) {  
			GetModuleFileName (GetWindowWord (hWndTarg, GWW_HINSTANCE),
			                   szWinText, sizeof (szWinText));
			return SetToken (ptDest, TTYPE_STR, 0, (DWORD)(LPSTR) szWinText);
		}	
	} else
		return ERR_TOO_FEW_PARMS;
	return SetToken (ptDest, TTYPE_NUM, 0, (DWORD)MYFALSE);
}  	
//------------------------------------------------------------
// WinMsgBox - Predefined func that displays a message box
// on the screen.
// Arg 1 - Msg box title 
// Arg 2 - Msg box text
// Arg 3 - Button/icon flags
//------------------------------------------------------------ 
INT WinMsgBox (LPLINETOKEN *pltLine, PTOKEN ptDest) {
	PTOKEN	ptArg;
	INT		sReturn;
	
	ptArg = (PTOKEN) ptDest->dwData;
	if (ptDest->wData < 4) 
		return ERR_TOO_FEW_PARMS;
 	
	sReturn = MessageBox (hMain, GetTokenString (ptArg-1),
	                      GetTokenString (ptArg-2), 
	                      (UINT) GetTokenVal (ptArg-3));
	return SetToken (ptDest, TTYPE_NUM, 0, (DWORD)sReturn);
}  	
//------------------------------------------------------------
// WinAskBox - Predefined func that displays a dialog box
// with an input field.
// Arg 1 - Win text
// Arg 2 - default answer
//------------------------------------------------------------ 
INT WinAskBox (LPLINETOKEN *pltLine, PTOKEN ptDest) {
	char		szWinText[MAXLINELEN];
	PTOKEN	ptArg;
	INT		sRet;
	
	ptArg = (PTOKEN) ptDest->dwData;
	if (ptDest->wData < 2) 
		return ERR_TOO_FEW_PARMS;
 
	szWinText[0] = 0;
	pszAskAns = szWinText;
	sRet = MyDisplayDialog(hInst, "AskBox", hMain, (WNDPROC)AskDlgProc,
	                MAKELONG (GetTokenString (ptArg-1),
	                          GetTokenString (ptArg-2)));
	if (sRet)
		return SetToken (ptDest, TTYPE_STR, 0, (DWORD)(LPSTR) szWinText);
	else	
		return SetToken (ptDest, TTYPE_STR, 0, (DWORD)(LPSTR)szNull);
}  	
//------------------------------------------------------------
// WinTicks - Predefined func that returns the number of
// ticks since windows started.
// No Args
//------------------------------------------------------------ 
INT WinTicks (LPLINETOKEN *pltLine, PTOKEN ptDest) {
	return SetToken (ptDest, TTYPE_NUM, 0, GetTickCount());
}
//------------------------------------------------------------
// WinGetEnvVar - Predefined func that returns an environment 
// variable.
// Arg 1 - Env Var name
//------------------------------------------------------------ 
INT WinGetEnvVar (LPLINETOKEN *pltLine, PTOKEN ptDest) {
	char		*pszVar;
	PTOKEN	ptArg;
	
	ptArg = (PTOKEN) ptDest->dwData;
	if (ptDest->wData < 2) 
		return ERR_TOO_FEW_PARMS;

	pszVar = getenv (GetTokenString (ptArg-1));
	if (pszVar) 
		return SetToken (ptDest, TTYPE_STR, 0, (DWORD)(LPSTR) pszVar);
	else	
		return SetToken (ptDest, TTYPE_STR, 0, (DWORD)(LPSTR)szNull);
}  	
//------------------------------------------------------------
// WinGetDropFiles - Predefined func that returns the filename 
// of a dropped file.
// Arg 1 - file number to return
//------------------------------------------------------------ 
INT WinGetDropFile (LPLINETOKEN *pltLine, PTOKEN ptDest) {
	PTOKEN	ptArg;
	LPSTR		lpData;
	INT		i, sCnt, sIndex;
	
	ptArg = (PTOKEN) ptDest->dwData;
	if (ptDest->wData < 2) 
		return ERR_TOO_FEW_PARMS;

	if (hDropData == 0)
		return SetToken (ptDest, TTYPE_NUM, 0, (DWORD)-1);

	sIndex = (INT)GetTokenVal (ptArg-1);
	
	lpData = GlobalLock (hDropData);
	//Get number of files dropped.
	sCnt = *((LPINT)lpData)++;
	if (sCnt <= sIndex) {
		GlobalUnlock (hDropData);
		return SetToken (ptDest, TTYPE_NUM, 0, (DWORD)-1);
	}
	//Scan to correct filename		
	for (i = 0; i < sIndex; i++) {
		while (*lpData++ != '\0');
		lpData++;
	}		
	i = SetToken (ptDest, TTYPE_STR, 0, (DWORD)lpData);
	GlobalUnlock (hDropData);
	return i;
}  	
//------------------------------------------------------------
// WinGetDropCount - Predefined func that returns the number
// of dropped files.
//------------------------------------------------------------ 
INT WinGetDropCount (LPLINETOKEN *pltLine, PTOKEN ptDest) {
	LPSTR		lpData;
	INT		sRC;

	if (hDropData == 0)
		return SetToken (ptDest, TTYPE_NUM, 0, (DWORD)0);
	
	lpData = GlobalLock (hDropData);
	//Return number of files dropped.
	sRC =	SetToken (ptDest, TTYPE_NUM, 0, (DWORD)*(LPINT)lpData);
	GlobalUnlock (hDropData);
	return sRC;
}  	
